# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Function to cluster spots in point cloud and detect relevant aggregated
structures.
"""

import numpy as np

import apifish.stack as stack

from .utils import convert_spot_coordinates

from sklearn.cluster import DBSCAN


# ### Detect clusters ###

def detect_clusters(spots, voxel_size, radius=350, nb_min_spots=4):
    """Cluster spots and detect relevant aggregated structures.

    #. If two spots are distant within a specific radius, we consider they are
       related to each other.
    #. A minimum number spots related to each others defines a cluster.

    Parameters
    ----------
    spots : np.ndarray
        Coordinates of the detected spots with shape (nb_spots, 3) or
        (nb_spots, 2).
    voxel_size : int, float, Tuple(int, float) or List(int, float)
        Size of a voxel, in nanometer. One value per spatial dimension (zyx or
        yx dimensions). If it's a scalar, the same value is applied to every
        dimensions.
    radius : int
        The maximum distance between two samples for one to be considered as
        in the neighborhood of the other. Radius expressed in nanometer.
    nb_min_spots : int
        The number of spots in a neighborhood for a point to be considered as
        a core point (from which a cluster is expanded). This includes the
        point itself.

    Returns
    -------
    clustered_spots : np.ndarray
        Coordinates of the detected spots with shape (nb_spots, 4) or
        (nb_spots, 3). One coordinate per dimension (zyx or yx coordinates)
        plus the index of the cluster assigned to the spot. If no cluster was
        assigned, value is -1.
    clusters : np.ndarray
        Array with shape (nb_clusters, 5) or (nb_clusters, 4). One coordinate
        per dimension for the clusters centroid (zyx or yx coordinates), the
        number of spots detected in the clusters and its index.

    """
    # TODO check that the behavior is the same with float64 and int64
    #  coordinates
    # check parameters
    stack.check_array(spots, ndim=2, dtype=[np.float64, np.int64])
    stack.check_parameter(
        voxel_size=(int, float, tuple, list),
        radius=int,
        nb_min_spots=int)

    # check consistency between parameters
    dtype = spots.dtype
    ndim = spots.shape[1]
    if ndim not in [2, 3]:
        raise ValueError("Spot coordinates should be in 2 or 3 dimensions, "
                         "not {0}.".format(ndim))
    if isinstance(voxel_size, (tuple, list)):
        if len(voxel_size) != ndim:
            raise ValueError(
                "'voxel_size' must be a scalar or a sequence with {0} "
                "elements.".format(ndim))
    else:
        voxel_size = (voxel_size,) * ndim

    # case where no spot were detected
    if spots.size == 0:
        clustered_spots = np.array([], dtype=dtype).reshape((0, ndim + 1))
        clusters = np.array([], dtype=dtype).reshape((0, ndim + 2))
        return clustered_spots, clusters

    # cluster spots
    clustered_spots = _cluster_spots(
        spots, voxel_size, radius, nb_min_spots)

    # extract and shape clusters information
    clusters = _extract_information(clustered_spots)

    return clustered_spots, clusters


def _cluster_spots(spots, voxel_size, radius, nb_min_spots):
    """Assign a cluster to each spot.

    Parameters
    ----------
    spots : np.ndarray
        Coordinates of the detected spots with shape (nb_spots, 3) or
        (nb_spots, 2).
    voxel_size : Tuple(int, float) or List(int, float)
        Size of a voxel, in nanometer. One value per spatial dimension (zyx or
        yx dimensions).
    radius : int
        The maximum distance between two samples for one to be considered as
        in the neighborhood of the other. Radius expressed in nanometer.
    nb_min_spots : int
        The number of spots in a neighborhood for a point to be considered as
        a core point (from which a cluster is expanded). This includes the
        point itself.

    Returns
    -------
    clustered_spots : np.ndarray
        Coordinates of the detected spots with shape (nb_spots, 4) or
        (nb_spots, 3). One coordinate per dimension (zyx or yx coordinates)
        plus the index of the cluster assigned to the spot. If no cluster was
        assigned, value is -1.

    """
    # convert spots coordinates in nanometer
    spots_nanometer = convert_spot_coordinates(
        spots=spots, voxel_size=voxel_size)

    # fit a DBSCAN clustering algorithm with a specific radius
    dbscan = DBSCAN(eps=radius, min_samples=nb_min_spots)
    dbscan.fit(spots_nanometer)
    labels = dbscan.labels_
    labels = labels[:, np.newaxis]

    # assign a cluster to each spot if possible
    clustered_spots = spots.copy()
    clustered_spots = np.concatenate((clustered_spots, labels), axis=1)

    return clustered_spots


def _extract_information(clustered_spots):
    """Extract clusters information from clustered spots.

    Parameters
    ----------
    clustered_spots : np.ndarray
        Coordinates of the detected spots with shape (nb_spots, 4) or
        (nb_spots, 3). One coordinate per dimension (zyx or yx coordinates)
        plus the index of the cluster assigned to the spot. If no cluster was
        assigned, value is -1.

    Returns
    -------
    clusters : np.ndarray
        Array with shape (nb_clusters, 5) or (nb_clusters, 4). One coordinate
        per dimension for the cluster centroid (zyx or yx coordinates), the
        number of spots detected in the cluster and its index.

    """
    # extract information for 3-d cluster...
    if clustered_spots.shape[1] == 4:

        # get 3-d cluster labels
        labels_clusters = np.unique(
            clustered_spots[clustered_spots[:, 3] != -1, 3])
        if labels_clusters.size == 0:
            clusters = np.array([], dtype=np.int64).reshape((0, 5))
            return clusters

        # shape information
        clusters = []
        for label in labels_clusters:
            spots_in_cluster = clustered_spots[clustered_spots[:, 3] == label,
                                               :3]
            z_cluster, y_cluster, x_cluster = spots_in_cluster.mean(axis=0)
            nb_spots_cluster = len(spots_in_cluster)
            clusters.append([z_cluster, y_cluster, x_cluster,
                             nb_spots_cluster, label])
        clusters = np.array(clusters, dtype=np.int64)

    # ... and 2-d cluster
    else:

        # get 2-d cluster labels
        labels_clusters = np.unique(
            clustered_spots[clustered_spots[:, 2] != -1, 2])
        if labels_clusters.size == 0:
            clusters = np.array([], dtype=np.int64).reshape((0, 4))
            return clusters

        # shape information
        clusters = []
        for label in labels_clusters:
            spots_in_cluster = clustered_spots[clustered_spots[:, 2] == label,
                                               :2]
            y_cluster, x_cluster = spots_in_cluster.mean(axis=0)
            nb_spots_cluster = len(spots_in_cluster)
            clusters.append([y_cluster, x_cluster, nb_spots_cluster, label])
        clusters = np.array(clusters, dtype=np.int64)

    return clusters



# -*- coding: utf-8 -*-

"""
The apifish.deep_learning subpackage includes deep learning models and their
routines.
"""

import warnings

# check tensorflow package
try:
    import tensorflow
    tf_version = tensorflow.__version__
    if tf_version != "2.3.0":
        warnings.warn("Available tensorflow version ({0}) differs from "
                      "requirements (2.3.0).".format(tf_version),
                      category=ImportWarning)
except ImportError:
    raise ImportError(
        "Tensorflow package (v2.3.0) is missing. You can install it by "
        "running the command 'pip install tensorflow==2.3.0' in a bash shell.")

# check tensorflow-addons package
try:
    import tensorflow_addons
    tfa_version = tensorflow_addons.__version__
    if tfa_version != "0.12.1":
        warnings.warn("Available tensorflow-addons version ({0}) differs from "
                      "requirements (0.12.1).".format(tfa_version),
                      category=ImportWarning)
except ImportError:
    raise ImportError(
        "Tensorflow addons package (v0.12.1) is missing. You can install it by"
        "running the command 'pip install tensorflow-addons==0.12.1' in a "
        "bash shell.")

from .utils_models import SameConv
from .utils_models import UpConv
from .utils_models import DownBlock
from .utils_models import UpBlock
from .utils_models import Encoder
from .utils_models import Decoder
from .utils_models import EncoderDecoder

from .models_segmentation import load_pretrained_model
from .models_segmentation import check_pretrained_weights
from .models_segmentation import build_compile_3_classes_model
from .models_segmentation import build_compile_distance_model
from .models_segmentation import build_compile_double_distance_model


_utils_models = [
    "SameConv",
    "DownBlock",
    "UpBlock",
    "Encoder",
    "Decoder",
    "EncoderDecoder"]

_models_segmentation = [
    "load_pretrained_model",
    "check_pretrained_weights",
    "build_compile_3_classes_model",
    "build_compile_distance_model",
    "build_compile_double_distance_model"]


__all__ = (_utils_models, _models_segmentation)



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Functions to model spots by fitting gaussian parameters.
"""

import numpy as np

import apifish.stack as stack

from .utils import _get_spot_volume
from .utils import _get_spot_surface
from .utils import get_object_radius_pixel

from scipy.special import erf
from scipy.optimize import curve_fit


# ### Spot modelization ###

def modelize_spot(reference_spot, voxel_size, spot_radius, return_coord=False):
    """Fit a gaussian function on the reference spot.

    Parameters
    ----------
    reference_spot : np.ndarray
        A 3-d or 2-d image with detected spot and shape (z, y, x) or (y, x).
    voxel_size : int, float, Tuple(int, float) or List(int, float)
        Size of a voxel, in nanometer. One value per spatial dimension (zyx or
        yx dimensions). If it's a scalar, the same value is applied to every
        dimensions.
    spot_radius : int, float, Tuple(int, float) or List(int, float)
        Radius of the spot, in nanometer. One value per spatial dimension (zyx
        or yx dimensions). If it's a scalar, the same radius is applied to
        every dimensions.
    return_coord : bool
        Return gaussian coordinates.

    Returns
    -------
    parameters_fitted : Tuple[float]
        * mu_z : float (optional)
            Coordinate of the gaussian center along the z axis, in pixel.
        * mu_y : float (optional)
            Coordinate of the gaussian center along the y axis, in pixel.
        * mu_x : float (optional)
            Coordinate of the gaussian center along the x axis, in pixel.
        * sigma_z : float
            Standard deviation of the gaussian along the z axis, in pixel.
            Available only for a 3-d modelization.
        * sigma_yx : float
            Standard deviation of the gaussian in the yx plan, in pixel.
        * amplitude : float
            Amplitude of the gaussian.
        * background : float
            Background minimum value of the image.

    """
    # check parameters
    stack.check_array(
        reference_spot,
        ndim=[2, 3],
        dtype=[np.uint8, np.uint16, np.float32, np.float64])
    stack.check_parameter(
        voxel_size=(int, float, tuple, list),
        spot_radius=(int, float, tuple, list),
        return_coord=bool)

    # check consistency between parameters
    ndim = reference_spot.ndim
    if isinstance(voxel_size, (tuple, list)):
        if len(voxel_size) != ndim:
            raise ValueError(
                "'voxel_size' must be a scalar or a sequence with {0} "
                "elements.".format(ndim))
    else:
        voxel_size = (voxel_size,) * ndim
    if isinstance(spot_radius, (tuple, list)):
        if len(spot_radius) != ndim:
            raise ValueError(
                "'spot_radius' must be a scalar or a sequence with {0} "
                "elements.".format(ndim))
    else:
        spot_radius = (spot_radius,) * ndim

    # initialize a grid representing the reference spot
    grid, centroid_coord = initialize_grid(
        image_spot=reference_spot,
        voxel_size=voxel_size,
        return_centroid=True)

    # compute amplitude and background of the reference spot
    amplitude, background = _initialize_background_amplitude(reference_spot)

    # initialize parameters of the gaussian function
    f = _objective_function(
        ndim=ndim,
        voxel_size=voxel_size,
        sigma_z=None,
        sigma_yx=None,
        amplitude=None)
    if ndim == 3:
        # parameters to fit: mu_z, mu_y, mu_x, sigma_z, sigma_yx, amplitude
        # and background
        centroid_z, centroid_y, centroid_x = centroid_coord
        p0 = [centroid_z, centroid_y, centroid_x, spot_radius[0],
              spot_radius[-1], amplitude, background]
        l_bound = [-np.inf, -np.inf, -np.inf, -np.inf, -np.inf, -np.inf, 0]
        u_bound = [np.inf, np.inf, np.inf, np.inf, np.inf, np.inf, np.inf]

    else:
        # parameters to fit: mu_y, mu_x, sigma_yx, amplitude and background
        centroid_y, centroid_x = centroid_coord
        p0 = [centroid_y, centroid_x, spot_radius[-1], amplitude, background]
        l_bound = [-np.inf, -np.inf, -np.inf, -np.inf, 0]
        u_bound = [np.inf, np.inf, np.inf, np.inf, np.inf]

    # fit a gaussian function on this reference spot
    popt, pcov = _fit_gaussian(
        f, grid, reference_spot, p0, lower_bound=l_bound, upper_bound=u_bound)

    # get optimized parameters to modelize the reference spot as a gaussian
    if ndim == 3:
        mu_z = popt[0]
        mu_y = popt[1]
        mu_x = popt[2]
        sigma_z = popt[3]
        sigma_yx = popt[4]
        amplitude = popt[5]
        background = popt[6]

        if return_coord:
            return mu_z, mu_y, mu_x, sigma_z, sigma_yx, amplitude, background
        else:
            return sigma_z, sigma_yx, amplitude, background

    else:
        mu_y = popt[0]
        mu_x = popt[1]
        sigma_yx = popt[2]
        amplitude = popt[3]
        background = popt[4]

        if return_coord:
            return mu_y, mu_x, sigma_yx, amplitude, background
        else:
            return sigma_yx, amplitude, background


def initialize_grid(image_spot, voxel_size, return_centroid=False):
    """Build a grid in nanometer to compute gaussian function values over a
    full volume or surface.

    Parameters
    ----------
    image_spot : np.ndarray
        An image with detected spot and shape (z, y, x) or (y, x).
    voxel_size : int, float, Tuple(int, float) or List(int, float)
        Size of a voxel, in nanometer. One value per spatial dimension (zyx or
        yx dimensions). If it's a scalar, the same value is applied to every
        dimensions.
    return_centroid : bool
        Compute centroid estimation of the grid.

    Returns
    -------
    grid : np.ndarray, np.float32
        A grid with the shape (3, z * y * x) or (2, y * x), in nanometer.
    centroid_coord : Tuple[float]
        Estimated centroid of the spot, in nanometer. One element per
        dimension.

    """
    # check parameters
    stack.check_array(
        image_spot,
        ndim=[2, 3],
        dtype=[np.uint8, np.uint16, np.float32, np.float64])
    stack.check_parameter(
        voxel_size=(int, float, tuple, list),
        return_centroid=bool)

    # check consistency between parameters
    ndim = image_spot.ndim
    if isinstance(voxel_size, (tuple, list)):
        if len(voxel_size) != ndim:
            raise ValueError(
                "'voxel_size' must be a scalar or a sequence with {0} "
                "elements.".format(ndim))
    else:
        voxel_size = (voxel_size,) * ndim

    # initialize grid in 2-d...
    if image_spot.ndim == 2:
        if return_centroid:
            grid, centroid_y, centroid_x = _initialize_grid_2d(
                image_spot, voxel_size[-1], return_centroid)
            return grid, (centroid_y, centroid_x)
        else:
            grid = _initialize_grid_2d(
                image_spot, voxel_size[-1], return_centroid)
            return grid

    # ... or 3-d
    else:
        if return_centroid:
            grid, centroid_z, centroid_y, centroid_x = _initialize_grid_3d(
                image_spot, voxel_size[0], voxel_size[-1], return_centroid)
            return grid, (centroid_z, centroid_y, centroid_x)
        else:
            grid = _initialize_grid_3d(
                image_spot, voxel_size[0], voxel_size[-1], return_centroid)
            return grid


def _initialize_grid_3d(image_spot, voxel_size_z, voxel_size_yx,
                        return_centroid=False):
    """Build a grid in nanometer to compute gaussian function values over a
    full volume.

    Parameters
    ----------
    image_spot : np.ndarray
        A 3-d image with detected spot and shape (z, y, x).
    voxel_size_z : int or float
        Size of a voxel along the z axis, in nanometer.
    voxel_size_yx : int or float
        Size of a voxel in the yx plan, in nanometer.
    return_centroid : bool
        Compute centroid estimation of the grid.

    Returns
    -------
    grid : np.ndarray, np.float32
        A grid with the shape (3, z * y * x), in nanometer.
    centroid_z : float
        Estimated centroid of the spot, in nanometer, along the z axis.
    centroid_y : float
        Estimated centroid of the spot, in nanometer, along the y axis.
    centroid_x : float
        Estimated centroid of the spot, in nanometer, along the x axis.

    """
    # get targeted size
    nb_z, nb_y, nb_x = image_spot.shape
    nb_pixels = image_spot.size

    # build meshgrid
    zz, yy, xx = np.meshgrid(np.arange(nb_z), np.arange(nb_y), np.arange(nb_x),
                             indexing="ij")
    zz = zz.astype(np.float32) * float(voxel_size_z)
    yy = yy.astype(np.float32) * float(voxel_size_yx)
    xx = xx.astype(np.float32) * float(voxel_size_yx)

    # format result
    grid = np.zeros((3, nb_pixels), dtype=np.float32)
    grid[0] = np.reshape(zz, (1, nb_pixels))
    grid[1] = np.reshape(yy, (1, nb_pixels))
    grid[2] = np.reshape(xx, (1, nb_pixels))

    # compute centroid of the grid
    if return_centroid:
        area = np.sum(image_spot)
        dz = image_spot * zz
        dy = image_spot * yy
        dx = image_spot * xx
        centroid_z = np.sum(dz) / area
        centroid_y = np.sum(dy) / area
        centroid_x = np.sum(dx) / area
        return grid, centroid_z, centroid_y, centroid_x

    else:
        return grid


def _initialize_grid_2d(image_spot, voxel_size_yx, return_centroid=False):
    """Build a grid in nanometer to compute gaussian function values over a
    full surface.

    Parameters
    ----------
    image_spot : np.ndarray
        A 2-d image with detected spot and shape (y, x).
    voxel_size_yx : int or float
        Size of a voxel in the yx plan, in nanometer.
    return_centroid : bool
        Compute centroid estimation of the grid.

    Returns
    -------
    grid : np.ndarray, np.float32
        A grid with the shape (2, y * x), in nanometer.
    centroid_y : float
        Estimated centroid of the spot along the y axis, in nanometer.
    centroid_x : float
        Estimated centroid of the spot along the x axis, in nanometer.

    """
    # get targeted size
    nb_y, nb_x = image_spot.shape
    nb_pixels = image_spot.size

    # build meshgrid
    yy, xx = np.meshgrid(np.arange(nb_y), np.arange(nb_x), indexing="ij")
    yy = yy.astype(np.float32) * float(voxel_size_yx)
    xx = xx.astype(np.float32) * float(voxel_size_yx)

    # format result
    grid = np.zeros((2, nb_pixels), dtype=np.float32)
    grid[0] = np.reshape(yy, (1, nb_pixels))
    grid[1] = np.reshape(xx, (1, nb_pixels))

    # compute centroid of the grid
    if return_centroid:
        area = np.sum(image_spot)
        dy = image_spot * yy
        dx = image_spot * xx
        centroid_y = np.sum(dy) / area
        centroid_x = np.sum(dx) / area
        return grid, centroid_y, centroid_x

    else:
        return grid


def _initialize_background_amplitude(image_spot):
    """Compute amplitude of a spot and background minimum value.

    Parameters
    ----------
    image_spot : np.ndarray, np.uint
        Image with detected spot and shape (z, y, x) or (y, x).

    Returns
    -------
    amplitude : float
        Amplitude of the spot.
    background : float
        Background minimum value of the voxel.

    """
    # compute values
    image_min, image_max = image_spot.min(), image_spot.max()
    amplitude = image_max - image_min
    background = image_min

    return amplitude, background


# ### Pixel fitting ###

def _objective_function(ndim, voxel_size, sigma_z, sigma_yx,
                        amplitude):
    """Design the objective function used to fit the gaussian function.

    Parameters
    ----------
    ndim : int
        Number of dimensions to consider (2 or 3).
    voxel_size : Tuple(int, float)
        Size of a voxel, in nanometer. One value per spatial dimension (zyx or
        yx dimensions).
    sigma_z : int, float or None
        Standard deviation of the gaussian along the z axis, in nanometer. If
        None, we consider a 2-d gaussian function.
    sigma_yx : int, float or None
        Standard deviation of the gaussian in the yx plan, in nanometer.
    amplitude : int, float or None
        Amplitude of the gaussian.

    Returns
    -------
    f : func
        A 3-d or 2-d gaussian function with some parameters fixed.

    """
    # define objective gaussian function
    if ndim == 3:
        f = _objective_function_3d(
            voxel_size_z=voxel_size[0],
            voxel_size_yx=voxel_size[-1],
            sigma_z=sigma_z,
            sigma_yx=sigma_yx,
            amplitude=amplitude)
    else:
        f = _objective_function_2d(
            voxel_size_yx=voxel_size[-1],
            sigma_yx=sigma_yx,
            amplitude=amplitude)

    return f


def _objective_function_3d(voxel_size_z, voxel_size_yx, sigma_z, sigma_yx,
                           amplitude):
    """Design the objective function used to fit the gaussian function.

    Parameters
    ----------
    voxel_size_z : int or float
        Size of a voxel along the z axis, in nanometer.
    voxel_size_yx : int or float
        Size of a voxel in the yx plan, in nanometer.
    sigma_z : int, float or None
        Standard deviation of the gaussian along the z axis, in nanometer.
    sigma_yx : int, float or None
        Standard deviation of the gaussian in the yx plan, in nanometer.
    amplitude : int, float or None
        Amplitude of the gaussian.

    Returns
    -------
    f : func
        A 3-d gaussian function with some parameters fixed.

    """
    # sigma is known, we fit mu, amplitude and background
    if (sigma_z is not None
            and sigma_yx is not None
            and amplitude is None):
        def f(grid, mu_z, mu_y, mu_x, amplitude, background):
            values = gaussian_3d(
                grid=grid,
                mu_z=mu_z,
                mu_y=mu_y,
                mu_x=mu_x,
                sigma_z=sigma_z,
                sigma_yx=sigma_yx,
                voxel_size_z=voxel_size_z,
                voxel_size_yx=voxel_size_yx,
                amplitude=amplitude,
                background=background)
            return values

    # amplitude is known, we fit sigma, mu and background
    elif (amplitude is not None
          and sigma_z is None
          and sigma_yx is None):
        def f(grid, mu_z, mu_y, mu_x, sigma_z, sigma_yx, background):
            values = gaussian_3d(
                grid=grid,
                mu_z=mu_z,
                mu_y=mu_y,
                mu_x=mu_x,
                sigma_z=sigma_z,
                sigma_yx=sigma_yx,
                voxel_size_z=voxel_size_z,
                voxel_size_yx=voxel_size_yx,
                amplitude=amplitude,
                background=background)
            return values

    # amplitude and sigma are known, we fit mu and background
    elif (amplitude is not None
          and sigma_z is not None
          and sigma_yx is not None):
        def f(grid, mu_z, mu_y, mu_x, background):
            values = gaussian_3d(
                grid=grid,
                mu_z=mu_z,
                mu_y=mu_y,
                mu_x=mu_x,
                sigma_z=sigma_z,
                sigma_yx=sigma_yx,
                voxel_size_z=voxel_size_z,
                voxel_size_yx=voxel_size_yx,
                amplitude=amplitude,
                background=background)
            return values

    # we fit mu, sigma, amplitude and background
    elif (amplitude is None
          and sigma_z is None
          and sigma_yx is None):
        def f(grid, mu_z, mu_y, mu_x, sigma_z, sigma_yx, amplitude,
              background):
            values = gaussian_3d(
                grid=grid,
                mu_z=mu_z,
                mu_y=mu_y,
                mu_x=mu_x,
                sigma_z=sigma_z,
                sigma_yx=sigma_yx,
                voxel_size_z=voxel_size_z,
                voxel_size_yx=voxel_size_yx,
                amplitude=amplitude,
                background=background)
            return values

    else:
        raise ValueError("Parameters 'sigma_z' and 'sigma_yx' should be set "
                         "or optimized together.")

    return f


# TODO add equations in the docstring
def gaussian_3d(grid, mu_z, mu_y, mu_x, sigma_z, sigma_yx, voxel_size_z,
                voxel_size_yx, amplitude, background,
                precomputed=None):
    """Compute the gaussian function over the grid representing a volume V
    with shape (V_z, V_y, V_x).

    Parameters
    ----------
    grid : np.ndarray, np.float
        Grid data to compute the gaussian function for different voxel within
        a volume V. In nanometer, with shape (3, V_z * V_y * V_x).
    mu_z : float
        Estimated mean of the gaussian signal along z axis, in nanometer.
    mu_y : float
        Estimated mean of the gaussian signal along y axis, in nanometer.
    mu_x : float
        Estimated mean of the gaussian signal along x axis, in nanometer.
    sigma_z : int or float
        Standard deviation of the gaussian along the z axis, in nanometer.
    sigma_yx : int or float
        Standard deviation of the gaussian in the yx plan, in nanometer.
    voxel_size_z : int or float
        Size of a voxel along the z axis, in nanometer.
    voxel_size_yx : int or float
        Size of a voxel in the yx plan, in nanometer.
    amplitude : float
        Estimated pixel intensity of the gaussian signal.
    background : float
        Estimated pixel intensity of the background.
    precomputed : Tuple[np.ndarray] or None
        Tuple with tables of precomputed values for the erf, with shape
        (nb_value, 2). One table per dimension.

    Returns
    -------
    values : np.ndarray, np.float
        Value of each voxel within the volume V according to the 3-d gaussian
        parameters. Shape (V_z * V_y * V_x,).

    """
    # get grid data to design a volume V
    meshgrid_z = grid[0]
    meshgrid_y = grid[1]
    meshgrid_x = grid[2]

    # use precomputed tables
    if precomputed is not None:
        # get tables
        table_erf_z = precomputed[0]
        table_erf_y = precomputed[1]
        table_erf_x = precomputed[2]

        # get indices for the tables
        i_z = np.abs(meshgrid_z - mu_z).astype(np.int64)
        i_y = np.abs(meshgrid_y - mu_y).astype(np.int64)
        i_x = np.abs(meshgrid_x - mu_x).astype(np.int64)

        # get precomputed values
        voxel_integral_z = table_erf_z[i_z, 1]
        voxel_integral_y = table_erf_y[i_y, 1]
        voxel_integral_x = table_erf_x[i_x, 1]

    # compute erf value
    else:
        # get voxel coordinates
        meshgrid_z_minus = meshgrid_z - voxel_size_z / 2
        meshgrid_z_plus = meshgrid_z + voxel_size_z / 2
        meshgrid_y_minus = meshgrid_y - voxel_size_yx / 2
        meshgrid_y_plus = meshgrid_y + voxel_size_yx / 2
        meshgrid_x_minus = meshgrid_x - voxel_size_yx / 2
        meshgrid_x_plus = meshgrid_x + voxel_size_yx / 2

        # compute gaussian function for each voxel (i, j, k) of volume V
        voxel_integral_z = _rescaled_erf(
            low=meshgrid_z_minus,
            high=meshgrid_z_plus,
            mu=mu_z,
            sigma=sigma_z)
        voxel_integral_y = _rescaled_erf(
            low=meshgrid_y_minus,
            high=meshgrid_y_plus,
            mu=mu_y,
            sigma=sigma_yx)
        voxel_integral_x = _rescaled_erf(
            low=meshgrid_x_minus,
            high=meshgrid_x_plus,
            mu=mu_x,
            sigma=sigma_yx)

    # compute 3-d gaussian values
    factor = amplitude / (voxel_size_yx ** 2 * voxel_size_z)
    voxel_integral = voxel_integral_z * voxel_integral_y * voxel_integral_x
    values = background + factor * voxel_integral

    return values


def _objective_function_2d(voxel_size_yx, sigma_yx, amplitude):
    """Design the objective function used to fit a 2-d gaussian function.

    Parameters
    ----------
    voxel_size_yx : int or float
        Size of a voxel in the yx plan, in nanometer.
    sigma_yx : int, float or None
        Standard deviation of the gaussian in the yx plan, in nanometer.
    amplitude : int, float or None
        Amplitude of the gaussian.

    Returns
    -------
    f : func
        A 2-d gaussian function with some parameters fixed.

    """
    # sigma is known, we fit mu, amplitude and background
    if sigma_yx is not None and amplitude is None:
        def f(grid, mu_y, mu_x, amplitude, background):
            values = gaussian_2d(
                grid=grid,
                mu_y=mu_y,
                mu_x=mu_x,
                sigma_yx=sigma_yx,
                voxel_size_yx=voxel_size_yx,
                amplitude=amplitude,
                background=background)
            return values

    # amplitude is known, we fit sigma, mu and background
    elif amplitude is not None and sigma_yx is None:
        def f(grid, mu_y, mu_x, sigma_yx, background):
            values = gaussian_2d(
                grid=grid,
                mu_y=mu_y,
                mu_x=mu_x,
                sigma_yx=sigma_yx,
                voxel_size_yx=voxel_size_yx,
                amplitude=amplitude,
                background=background)
            return values

    # amplitude and sigma are known, we fit mu and background
    elif amplitude is not None and sigma_yx is not None:
        def f(grid, mu_y, mu_x, background):
            values = gaussian_2d(
                grid=grid,
                mu_y=mu_y,
                mu_x=mu_x,
                sigma_yx=sigma_yx,
                voxel_size_yx=voxel_size_yx,
                amplitude=amplitude,
                background=background)
            return values

    # we fit mu, sigma, amplitude and background
    else:
        def f(grid, mu_y, mu_x, sigma_yx, amplitude, background):
            values = gaussian_2d(
                grid=grid,
                mu_y=mu_y,
                mu_x=mu_x,
                sigma_yx=sigma_yx,
                voxel_size_yx=voxel_size_yx,
                amplitude=amplitude,
                background=background)
            return values

    return f


# TODO add equations in the docstring
def gaussian_2d(grid, mu_y, mu_x, sigma_yx, voxel_size_yx, amplitude,
                background, precomputed=None):
    """Compute the gaussian function over the grid representing a surface S
    with shape (S_y, S_x).

    Parameters
    ----------
    grid : np.ndarray, np.float
        Grid data to compute the gaussian function for different voxel within
        a surface S. In nanometer, with shape (2, S_y * S_x).
    mu_y : float
        Estimated mean of the gaussian signal along y axis, in nanometer.
    mu_x : float
        Estimated mean of the gaussian signal along x axis, in nanometer.
    sigma_yx : int or float
        Standard deviation of the gaussian in the yx plan, in nanometer.
    voxel_size_yx : int or float
        Size of a voxel in the yx plan, in nanometer.
    amplitude : float
        Estimated pixel intensity of the gaussian signal.
    background : float
        Estimated pixel intensity of the background.
    precomputed : Tuple[np.ndarray] or None
        Tuple with tables of precomputed values for the erf, with shape
        (nb_value, 2). One table per dimension.

    Returns
    -------
    values : np.ndarray, np.float
        Value of each voxel within the surface S according to the 2-d gaussian
        parameters. Shape (S_y * S_x,).

    """
    # get grid data to design a surface S
    meshgrid_y = grid[0]
    meshgrid_x = grid[1]

    # use precomputed tables
    if precomputed is not None:
        # get tables
        table_erf_y = precomputed[0]
        table_erf_x = precomputed[1]

        # get indices for the tables
        i_y = np.abs(meshgrid_y - mu_y).astype(np.int64)
        i_x = np.abs(meshgrid_x - mu_x).astype(np.int64)

        # get precomputed values
        voxel_integral_y = table_erf_y[i_y, 1]
        voxel_integral_x = table_erf_x[i_x, 1]

    # compute erf value
    else:
        # get voxel coordinates
        meshgrid_y_minus = meshgrid_y - voxel_size_yx / 2
        meshgrid_y_plus = meshgrid_y + voxel_size_yx / 2
        meshgrid_x_minus = meshgrid_x - voxel_size_yx / 2
        meshgrid_x_plus = meshgrid_x + voxel_size_yx / 2

        # compute gaussian function for each voxel (i, j) of surface S
        voxel_integral_y = _rescaled_erf(
            low=meshgrid_y_minus,
            high=meshgrid_y_plus,
            mu=mu_y,
            sigma=sigma_yx)
        voxel_integral_x = _rescaled_erf(
            low=meshgrid_x_minus,
            high=meshgrid_x_plus,
            mu=mu_x,
            sigma=sigma_yx)

    # compute 2-d gaussian values
    factor = amplitude / (voxel_size_yx ** 2)
    voxel_integral = voxel_integral_y * voxel_integral_x
    values = background + factor * voxel_integral

    return values


def _rescaled_erf(low, high, mu, sigma):
    """Rescaled the Error function along a specific axis.

    # TODO add equations

    Parameters
    ----------
    low : np.ndarray, np.float
        Lower bound of the voxel along a specific axis.
    high : np.ndarray, np.float
        Upper bound of the voxel along a specific axis.
    mu : int or float
        Estimated mean of the gaussian signal along a specific axis.
    sigma : int or float
        Estimated standard deviation of the gaussian signal along a specific
        axis.

    Returns
    -------
    rescaled_erf : np.ndarray, np.float
        Rescaled erf along a specific axis.

    """
    # compute erf and normalize it
    low_ = (low - mu) / (np.sqrt(2) * sigma)
    high_ = (high - mu) / (np.sqrt(2) * sigma)
    rescaled_erf = sigma * np.sqrt(np.pi / 2) * (erf(high_) - erf(low_))

    return rescaled_erf


def _fit_gaussian(f, grid, image_spot, p0, lower_bound=None, upper_bound=None):
    """Fit a gaussian function to a 3-d or 2-d image.

    # TODO add equations and algorithm

    Parameters
    ----------
    f : func
        A 3-d or 2-d gaussian function with some parameters fixed.
    grid : np.ndarray, np.float
        Grid data to compute the gaussian function for different voxel within
        a volume V or surface S. In nanometer, with shape (3, V_z * V_y * V_x),
        or (2, S_y * S_x).
    image_spot : np.ndarray, np.uint
        A 3-d or 2-d image with detected spot and shape (z, y, x) or (y, x).
    p0 : List
        List of parameters to estimate.
    lower_bound : List or None
        List of lower bound values for the different parameters.
    upper_bound : List or None
        List of upper bound values for the different parameters.

    Returns
    -------
    popt : np.ndarray
        Fitted parameters.
    pcov : np.ndarray
        Estimated covariance of 'popt'.

    """
    # TODO check that we do not fit a 2-d gaussian function to a 3-d image or
    #  the opposite

    # compute lower bound and upper bound
    if lower_bound is None:
        lower_bound = [-np.inf for _ in p0]
    if upper_bound is None:
        upper_bound = [np.inf for _ in p0]
    bounds = (lower_bound, upper_bound)

    # Apply non-linear least squares to fit a gaussian function to a 3-d image
    y = np.reshape(image_spot, (image_spot.size,)).astype(np.float32)
    popt, pcov = curve_fit(f=f, xdata=grid, ydata=y, p0=p0, bounds=bounds)

    return popt, pcov


def precompute_erf(ndim, voxel_size, sigma, max_grid=200):
    """Precompute different values for the erf with a nanometer resolution.

    Parameters
    ----------
    ndim : int
        Number of dimensions to consider (2 or 3).
    voxel_size : int, float, Tuple(int, float) or List(int, float)
        Size of a voxel, in nanometer. One value per spatial dimension (zyx or
        yx dimensions). If it's a scalar, the same value is applied to every
        dimensions.
    sigma : int, float, Tuple(int, float) or List(int, float)
        Standard deviation of the gaussian, in nanometer. One value per
        spatial dimension (zyx or yx dimensions). If it's a scalar, the same
        value is applied to every dimensions.
    max_grid : int
        Maximum size of the grid on which we precompute the erf, in pixel.

    Returns
    -------
    table_erf : Tuple[np.ndarray]
        Tuple with tables of precomputed values for the erf, with shape
        (nb_value, 2). One table per dimension. First column is the coordinate
        along the table dimension. Second column is the precomputed erf value.

    """
    # check parameters
    stack.check_parameter(
        ndim=int,
        voxel_size=(int, float, tuple, list),
        sigma=(int, float, tuple, list),
        max_grid=int)
    if ndim not in [2, 3]:
        raise ValueError(
            "Number of dimensions requested should be 2 or 3, "
            "not {0}.".format(ndim))

    # check consistency between parameters
    if isinstance(voxel_size, (tuple, list)):
        if len(voxel_size) != ndim:
            raise ValueError(
                "'voxel_size' must be a scalar or a sequence with {0} "
                "elements.".format(ndim))
    else:
        voxel_size = (voxel_size,) * ndim
    if isinstance(sigma, (tuple, list)):
        if len(sigma) != ndim:
            raise ValueError(
                "'sigma' must be a scalar or a sequence with {0} "
                "elements.".format(ndim))
    else:
        sigma = (sigma,) * ndim

    # build a grid with a spatial resolution of 1 nm and a size of
    # max_grid * resolution nm
    max_size_yx = np.ceil(max_grid * voxel_size[-1]).astype(np.int64)
    yy = np.array([i for i in range(0, max_size_yx)])
    xx = np.array([i for i in range(0, max_size_yx)])
    mu_y, mu_x = 0, 0

    # compute erf values for this grid
    erf_y = _rescaled_erf(
        low=yy - voxel_size[-1] / 2,
        high=yy + voxel_size[-1] / 2,
        mu=mu_y,
        sigma=sigma[-1])
    erf_x = _rescaled_erf(
        low=xx - voxel_size[-1] / 2,
        high=xx + voxel_size[-1] / 2,
        mu=mu_x,
        sigma=sigma[-1])
    table_erf_y = np.array([yy, erf_y]).T
    table_erf_x = np.array([xx, erf_x]).T

    # precompute erf along z axis if needed
    if ndim == 2:
        return table_erf_y, table_erf_x

    else:
        max_size_z = np.ceil(max_grid * voxel_size[0]).astype(np.int64)
        zz = np.array([i for i in range(0, max_size_z)])
        mu_z = 0
        erf_z = _rescaled_erf(
            low=zz - voxel_size[0] / 2,
            high=zz + voxel_size[0] / 2,
            mu=mu_z,
            sigma=sigma[0])
        table_erf_z = np.array([zz, erf_z]).T
        return table_erf_z, table_erf_y, table_erf_x


# ### Subpixel fitting ###

def fit_subpixel(image, spots, voxel_size, spot_radius):
    """Fit gaussian signal on every spot to find a subpixel coordinates.

    Parameters
    ----------
    image : np.ndarray
        Image with shape (z, y, x) or (y, x).
    spots : np.ndarray
        Coordinate of the spots detected, with shape (nb_spots, 3) or
        (nb_spots, 2). One coordinate per dimension (zyx or yx coordinates).
    voxel_size : int, float, Tuple(int, float) or List(int, float)
        Size of a voxel, in nanometer. One value per spatial dimension (zyx or
        yx dimensions). If it's a scalar, the same value is applied to every
        dimensions.
    spot_radius : int, float, Tuple(int, float) or List(int, float)
        Radius of the spot, in nanometer. One value per spatial dimension (zyx
        or yx dimensions). If it's a scalar, the same radius is applied to
        every dimensions.

    Returns
    -------
    spots_subpixel : np.ndarray, np.float64
        Coordinate of the spots detected, with shape (nb_spots, 3) or
        (nb_spots, 2). One coordinate per dimension (zyx or yx coordinates).

    """
    # check parameters
    stack.check_array(
        image,
        ndim=[2, 3],
        dtype=[np.uint8, np.uint16, np.float32, np.float64])
    stack.check_array(spots, ndim=2, dtype=[np.float64, np.int64])
    stack.check_parameter(
        voxel_size=(int, float, tuple, list),
        spot_radius=(int, float, tuple, list))

    # check consistency between parameters
    ndim = image.ndim
    if ndim != spots.shape[1]:
        raise ValueError("Provided image has {0} dimensions but spots are "
                         "detected in {1} dimensions."
                         .format(ndim, spots.shape[1]))
    if isinstance(voxel_size, (tuple, list)):
        if len(voxel_size) != ndim:
            raise ValueError(
                "'voxel_size' must be a scalar or a sequence with {0} "
                "elements.".format(ndim))
    else:
        voxel_size = (voxel_size,) * ndim
    if isinstance(spot_radius, (tuple, list)):
        if len(spot_radius) != ndim:
            raise ValueError(
                "'spot_radius' must be a scalar or a sequence with {0} "
                "elements.".format(ndim))
    else:
        spot_radius = (spot_radius,) * ndim

    # compute radius used to crop spot image
    radius_pixel = get_object_radius_pixel(
        voxel_size_nm=voxel_size,
        object_radius_nm=spot_radius,
        ndim=ndim)
    radius = [np.sqrt(ndim) * r for r in radius_pixel]
    radius = tuple(radius)

    # loop over every spot
    spots_subpixel = []
    for coord in spots[:, :ndim]:

        # fit subpixel coordinates
        if ndim == 3:
            subpixel_coord = _fit_subpixel_3d(
                image=image, coord=coord,
                radius_to_crop=radius,
                voxel_size_z=voxel_size[0],
                voxel_size_yx=voxel_size[-1],
                spot_radius_z=spot_radius[0],
                spot_radius_yx=spot_radius[-1])
        else:
            subpixel_coord = _fit_subpixel_2d(
                image=image, coord=coord,
                radius_to_crop=radius,
                voxel_size_yx=voxel_size[-1],
                spot_radius_yx=spot_radius[-1])
        spots_subpixel.append(subpixel_coord)

    # format results
    spots_subpixel = np.stack(spots_subpixel)

    return spots_subpixel


def _fit_subpixel_3d(image, coord, radius_to_crop, voxel_size_z, voxel_size_yx,
                     spot_radius_z, spot_radius_yx):
    """Fit a 3-d gaussian on a detected spot.

    Parameters
    ----------
    image : np.ndarray
        Image with shape (z, y, x).
    coord : np.ndarray, np.int64
        Coordinate of the spot detected, with shape (3,). One coordinate per
        dimension (zyx coordinates).
    radius_to_crop : Tuple[float]
        Enlarged radius of a spot, in pixel, used to crop an image around it.
        Tuple with 3 scalars (one per dimension zyx).
    voxel_size_z : int or float
        Size of a voxel along the z axis, in nanometer.
    voxel_size_yx : int or float
        Size of a voxel in the yx plan, in nanometer.
    spot_radius_z : int or float
        Radius of the spot along the z axis, in nanometer.
    spot_radius_yx : int or float
        Radius of the spot in the yx plan, in nanometer.

    Returns
    -------
    new_coord : List[float]
        Coordinates of the spot centroid with a subpixel accuracy (one element
        per dimension).

    """
    # extract spot image
    image_spot, bbox_low = _get_spot_volume(
        image=image,
        spot_z=coord[0],
        spot_y=coord[1],
        spot_x=coord[2],
        radius_z=radius_to_crop[0],
        radius_yx=radius_to_crop[-1])

    # fit gaussian
    try:
        parameters = modelize_spot(
            reference_spot=image_spot,
            voxel_size=(voxel_size_z, voxel_size_yx, voxel_size_yx),
            spot_radius=(spot_radius_z, spot_radius_yx, spot_radius_yx),
            return_coord=True)

        # format coordinates and ensure it is fitted within the spot image
        z_max, y_max, x_max = image_spot.shape
        coord_z = parameters[0] / voxel_size_z
        if coord_z < 0 or coord_z > z_max:
            coord_z = coord[0]
        else:
            coord_z += bbox_low[0]
        coord_y = parameters[1] / voxel_size_yx
        if coord_y < 0 or coord_y > y_max:
            coord_y = coord[1]
        else:
            coord_y += bbox_low[1]
        coord_x = parameters[2] / voxel_size_yx
        if coord_x < 0 or coord_x > x_max:
            coord_x = coord[2]
        else:
            coord_x += bbox_low[2]
        new_coord = [coord_z, coord_y, coord_x]

    # if a spot is ill-conditioned, we simply keep its original coordinates
    except RuntimeError:
        new_coord = list(coord)

    return new_coord


def _fit_subpixel_2d(image, coord, radius_to_crop, voxel_size_yx,
                     spot_radius_yx):
    """Fit a 2-d gaussian on a detected spot.

    Parameters
    ----------
    image : np.ndarray
        Image with shape (y, x).
    coord : np.ndarray, np.int64
        Coordinate of the spot detected, with shape (2,). One coordinate per
        dimension (yx coordinates).
    radius_to_crop : Tuple[float]
        Enlarged radius of a spot, in pixel, used to crop an image around it.
        Tuple with 2 scalars (one per dimension yx).
    voxel_size_yx : int or float
        Size of a voxel in the yx plan, in nanometer.
    spot_radius_yx : int or float
        Radius of the spot in the yx plan, in nanometer.

    Returns
    -------
    new_coord : List[float]
        Coordinates of the spot centroid with a subpixel accuracy (one element
        per dimension).

    """
    # extract spot image
    image_spot, bbox_low = _get_spot_surface(
        image=image,
        spot_y=coord[0],
        spot_x=coord[1],
        radius_yx=radius_to_crop[-1])

    # fit gaussian
    try:
        parameters = modelize_spot(
            reference_spot=image_spot,
            voxel_size=(voxel_size_yx, voxel_size_yx),
            spot_radius=(spot_radius_yx, spot_radius_yx),
            return_coord=True)

        # format coordinates and ensure it is fitted within the spot image
        y_max, x_max = image_spot.shape
        coord_y = parameters[0] / voxel_size_yx
        if coord_y < 0 or coord_y > y_max:
            coord_y = coord[0]
        else:
            coord_y += bbox_low[0]
        coord_x = parameters[1] / voxel_size_yx
        if coord_x < 0 or coord_x > x_max:
            coord_x = coord[1]
        else:
            coord_x += bbox_low[1]
        new_coord = [coord_y, coord_x]

    # if a spot is ill-conditioned, we simply keep its original coordinates
    except RuntimeError:
        new_coord = list(coord)

    return new_coord



apifishapifishapifish# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for apifish.detection.spot_modeling module.
"""


# TODO test apifish.detection.modelize_spot
# TODO test apifish.detection.initialize_grid
# TODO test apifish.detection.gaussian_2d
# TODO test apifish.detection.gaussian_3d
# TODO test apifish.detection.precompute_erf
# TODO test apifish.detection.fit_subpixel



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Class and functions to segment nucleus.
"""

import numpy as np

import apifish.stack as stack

from .postprocess import label_instances
from .postprocess import clean_segmentation

from skimage.morphology.selem import disk
from skimage.morphology import reconstruction


# ### Unet model ###

def unet_3_classes_nuc():
    """Load a pretrained Unet model to predict 3 classes from nucleus images:
    background, edge and foreground.

    Returns
    -------
    model : ``tensorflow.keras.model`` object
        Pretrained Unet model.

    """
    # import  deep_learning subpackage
    import apifish.deep_learning as dl

    # load model
    model = dl.load_pretrained_model("nuc", "3_classes")

    return model


def apply_unet_3_classes(model, image, target_size=None,
                         test_time_augmentation=False):
    """Segment image with a 3-classes trained model.

    Parameters
    ----------
     model : ``tensorflow.keras.model`` object
        Pretrained Unet model that predicts 3 classes from nucleus or cell
        images (background, edge and foreground).
    image : np.ndarray, np.uint
        Original image to segment with shape (y, x).
    target_size : int
        Resize image before segmentation. A squared image is resize to
        `target_size`. A rectangular image is resize such that its smaller
        dimension equals `target_size`.
    test_time_augmentation : bool
        Apply test time augmentation or not. The image is augmented 8 times
        and the final segmentation is the average result over these
        augmentations.

    Returns
    -------
    image_label_pred : np.ndarray, np.int64
        Labelled image. Each instance is characterized by the same pixel value.

    """
    # check parameters
    stack.check_parameter(
        target_size=(int, type(None)),
        test_time_augmentation=bool)
    stack.check_array(image, ndim=2, dtype=[np.uint8, np.uint16])

    # get original shape
    height, width = image.shape

    # resize image if necessary
    if target_size is None:
        # keep the original shape
        image_to_process = image.copy()
        new_height, new_width = height, width

    else:
        # target size should be multiple of 16
        target_size -= 16

        # we resize the images below the target size
        ratio = target_size / min(height, width)
        new_height = int(np.round(height * ratio))
        new_width = int(np.round(width * ratio))
        new_shape = (new_height, new_width)
        image_to_process = stack.resize_image(image, new_shape, "bilinear")

    # get padding marge to make it multiple of 16
    marge_padding = stack.get_marge_padding(new_height, new_width, x=16)
    top, bottom = marge_padding[0]
    left, right = marge_padding[1]
    image_to_process = np.pad(
        image_to_process, pad_width=marge_padding, mode='symmetric')

    # standardize and cast image
    image_to_process = stack.compute_image_standardization(image_to_process)
    image_to_process = image_to_process.astype(np.float32)

    # augment images
    if test_time_augmentation:
        image_to_process = stack.augment_8_times(image_to_process)
        n_augmentations = 8
    else:
        image_to_process = [image_to_process]
        n_augmentations = 1

    # loop over augmentations
    predictions_augmented = []
    for i in range(n_augmentations):

        # get images
        image_to_process_ = image_to_process[i]
        image_to_process_ = image_to_process_[np.newaxis, :, :, np.newaxis]

        # make predictions
        prediction = model.predict(image_to_process_)

        # remove padding
        if i in [0, 1, 2, 6]:
            prediction = prediction[0, top:-bottom, left:-right, :]
        else:
            prediction = prediction[0, left:-right, top:-bottom, :]

        # resize predictions back taking into account a potential deformation
        # from the image augmentation
        if target_size is not None:
            if i in [0, 1, 2, 6]:
                prediction = stack.resize_image(
                    prediction, (height, width), "bilinear")
            else:
                prediction = stack.resize_image(
                    prediction, (width, height), "bilinear")

        # store predictions
        predictions_augmented.append(prediction)

    # reversed image augmentation
    if test_time_augmentation:
        predictions = stack.augment_8_times_reversed(predictions_augmented)
        mean_prediction = np.mean(predictions, axis=0)
    else:
        mean_prediction = predictions_augmented[0]

    # postprocess predictions
    image_label_pred = from_3_classes_to_instances(mean_prediction)

    return image_label_pred


def from_3_classes_to_instances(label_3_classes):
    """Extract instance labels from 3-classes Unet output.

    Parameters
    ----------
    label_3_classes : np.ndarray, np.float32
        Model prediction about the nucleus surface and boundaries, with shape
        (y, x, 3).

    Returns
    -------
    label : np.ndarray, np.int64
        Labelled image. Each instance is characterized by the same pixel value.

    """
    # check parameters
    stack.check_array(label_3_classes, ndim=3, dtype=[np.float32])

    # get classes indices
    label_3_classes = np.argmax(label_3_classes, axis=-1)

    # keep foreground predictions
    mask = label_3_classes > 1

    # instantiate each individual foreground surface predicted
    label = label_instances(mask)

    # dilate label
    label = label.astype(np.float64)
    label = stack.dilation_filter(label, kernel_shape="disk", kernel_size=1)
    label = label.astype(np.int64)

    return label


# ### Utility functions ###

def remove_segmented_nuc(image, nuc_mask, size_nuclei=2000):
    """Remove the nuclei we have already segmented in an image.

    #. We start from the segmented nuclei with a light dilation. The missed
       nuclei and the background are set to 0 and removed from the original
       image.
    #. We reconstruct the missing nuclei by small dilation. As we used the
       original image to set the maximum allowed value at each pixel, the
       background pixels remain unchanged. However, pixels from the missing
       nuclei are partially reconstructed by the dilation. The reconstructed
       image only differs from the original one where the nuclei have been
       missed.
    #. We subtract the reconstructed image from the original one.
    #. From the few missing nuclei kept and restored, we build a binary mask
       (dilation, small object removal).
    #. We apply this mask to the original image to get the original pixel
       intensity of the missing nuclei.
    #. We remove pixels with a too low intensity.

    Parameters
    ----------
    image : np.ndarray, np.uint
        Original nuclei image with shape (y, x).
    nuc_mask : np.ndarray,
        Result of the segmentation (with instance differentiation or not).
    size_nuclei : int
        Threshold above which we detect a nuclei.

    Returns
    -------
    image_without_nuc : np.ndarray
        Image with shape (y, x) and the same dtype of the original image.
        Nuclei previously detected in the mask are removed.

    """
    # check parameters
    stack.check_array(image, ndim=2, dtype=[np.uint8, np.uint16])
    stack.check_array(nuc_mask, ndim=2, dtype=bool)
    stack.check_parameter(size_nuclei=int)

    # store original dtype
    original_dtype = image.dtype

    # dilate the mask
    mask_dilated = stack.dilation_filter(image, "disk", 10)

    # remove the unsegmented nuclei from the original image
    diff = image.copy()
    diff[mask_dilated == 0] = 0

    # reconstruct the missing nuclei by dilation
    s = disk(1).astype(original_dtype)
    image_reconstructed = reconstruction(diff, image, selem=s)
    image_reconstructed = image_reconstructed.astype(original_dtype)

    # subtract the reconstructed image from the original one
    image_filtered = image.copy()
    image_filtered -= image_reconstructed

    # build the binary mask for the missing nuclei
    missing_mask = image_filtered > 0
    missing_mask = clean_segmentation(
        missing_mask,
        small_object_size=size_nuclei,
        fill_holes=True)
    missing_mask = stack.dilation_filter(missing_mask, "disk", 20)

    # TODO improve the thresholds
    # get the original pixel intensity of the unsegmented nuclei
    unsegmented_nuclei = image.copy()
    unsegmented_nuclei[missing_mask == 0] = 0
    if original_dtype == np.uint8:
        unsegmented_nuclei[unsegmented_nuclei < 40] = 0
    else:
        unsegmented_nuclei[unsegmented_nuclei < 10000] = 0

    return unsegmented_nuclei



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for apifish.stack.utils module.
"""

import pytest

import apifish.stack as stack

import numpy as np
import pandas as pd


# TODO add test for apifish.stack.load_and_save_url
# TODO add test for apifish.stack.check_hash
# TODO add test for apifish.stack.compute_hash
# TODO add test for apifish.stack.check_input_data
# TODO add test for apifish.stack.moving_average
# TODO add test for apifish.stack.centered_moving_average

# ### Test sanity check functions ###

def test_check_parameter():
    # define a function with different parameters to check
    def foo(a, b, c, d, e, f, g, h):
        stack.check_parameter(a=(list, type(None)),
                              b=str,
                              c=int,
                              d=float,
                              e=np.ndarray,
                              f=bool,
                              g=(pd.DataFrame, pd.Series),
                              h=pd.DataFrame)
        return True

    # test the consistency of the check function when it works...
    assert foo(a=[], b="bar", c=5, d=2.5, e=np.array([3, 6, 9]),
               f=True, g=pd.DataFrame(), h=pd.DataFrame())
    assert foo(a=None, b="", c=10, d=2.0, e=np.array([3, 6, 9]),
               f=False, g=pd.Series(), h=pd.DataFrame())

    # ... and when it should raise an error
    with pytest.raises(TypeError):
        foo(a=(), b="bar", c=5, d=2.5, e=np.array([3, 6, 9]),
            f=True, g=pd.DataFrame(), h=pd.DataFrame())
    with pytest.raises(TypeError):
        foo(a=[], b="bar", c=5.0, d=2.5, e=np.array([3, 6, 9]),
            f=True, g=pd.DataFrame(), h=pd.DataFrame())
    with pytest.raises(TypeError):
        foo(a=[], b="bar", c=5, d=2, e=np.array([3, 6, 9]),
            f=True, g=pd.DataFrame(), h=pd.DataFrame())
    with pytest.raises(TypeError):
        foo(a=[], b="bar", c=5, d=2.5, e=[3, 6, 9],
            f=True, g=pd.DataFrame(), h=pd.DataFrame())
    with pytest.raises(TypeError):
        foo(a=[], b="bar", c=5, d=2.5, e=np.zeros((3, 3)),
            f=True, g=pd.DataFrame(), h=pd.Series())


def test_check_df():
    # build a dataframe to test
    df = pd.DataFrame({"A": [3, 6, 9],
                       "B": [2.5, np.nan, 1.3],
                       "C": ["arthur", "florian", "thomas"],
                       "D": [True, True, False]})

    # test the consistency of the check function when it works...
    assert stack.check_df(df,
                          features=["A", "B", "C", "D"],
                          features_without_nan=["A", "C", "D"])
    assert stack.check_df(df,
                          features=["B", "A"],
                          features_without_nan=["C", "D", "A"])
    assert stack.check_df(df,
                          features=None,
                          features_without_nan=["A", "C", "D"])
    assert stack.check_df(df,
                          features=["A", "B", "C", "D"],
                          features_without_nan=None)

    # ... and when it should raise an error
    with pytest.raises(ValueError):
        stack.check_df(df,
                       features=["A", "B", "C", "D", "E"],
                       features_without_nan=["A", "C", "D"])
    with pytest.raises(ValueError):
        stack.check_df(df,
                       features=["A", "B", "C", "D"],
                       features_without_nan=["A", "B", "C", "D"])


def test_check_array():
    # build some arrays to test
    a = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=np.float32)
    b = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=np.int64)
    c = np.array(([1, 2, 3]), dtype=np.float32)
    d = np.array(([1, 2, np.nan]), dtype=np.float32)

    # test number of dimensions
    assert stack.check_array(a, ndim=2)
    assert stack.check_array(b, ndim=[1, 2])
    assert stack.check_array(c, ndim=1)
    with pytest.raises(ValueError):
        stack.check_array(a, ndim=1)

    # test dtypes
    assert stack.check_array(a, dtype=np.float32)
    assert stack.check_array(b, dtype=np.int64)
    assert stack.check_array(c, dtype=[np.float32, np.int64])
    with pytest.raises(TypeError):
        stack.check_array(a, dtype=np.float64)

    # test missing values
    assert stack.check_array(a, allow_nan=False)
    assert stack.check_array(b, allow_nan=True)
    assert stack.check_array(d, allow_nan=True)
    with pytest.raises(ValueError):
        stack.check_array(d, allow_nan=False)


def test_check_range_value():
    # build some arrays to test
    a = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]], dtype=np.float32)

    # test the consistency of the check function when it works...
    assert stack.check_range_value(a, min_=1, max_=None)
    assert stack.check_range_value(a, min_=None, max_=9)

    # ... and when it should raise an error
    with pytest.raises(ValueError):
        stack.check_range_value(a, min_=2, max_=None)
    with pytest.raises(ValueError):
        stack.check_range_value(a, min_=None, max_=8)


# ### Constants ###

def test_margin_value():
    # test margin value
    assert stack.get_margin_value() >= 2


def test_epsilon_float_32():
    # test epsilon value and dtype
    eps = stack.get_eps_float32()
    assert eps < 1e-5
    assert isinstance(eps, np.float32)



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Functions used to read data from various sources and store them in a
multidimensional tensor (numpy.ndarray).
"""

import mrc
import warnings

import numpy as np
import pandas as pd

from skimage import io
from .utils import check_array
from .utils import check_parameter

from astropy.table import Table

# TODO add general read function with mime types
# TODO saving data in csv does not preserve dtypes

# ### Read ###

def read_image(path, sanity_check=False):
    """Read an image with ``png``, ``jpg``, ``jpeg``, ``tif`` or ``tiff``
    extension.

    Parameters
    ----------
    path : str
        Path of the image to read.
    sanity_check : bool
        Check if the array returned fits with apifish pipeline.

    Returns
    -------
    image : ndarray, np.uint or np.int
        Image read.

    """
    # check path
    check_parameter(
        path=str,
        sanity_check=bool)

    # read image
    image = io.imread(path)

    # check the output image
    if sanity_check:
        check_array(
            image,
            dtype=[np.uint8, np.uint16, np.uint32, np.uint64,
                   np.int8, np.int16, np.int32, np.int64,
                   np.float16, np.float32, np.float64,
                   bool],
            ndim=[2, 3, 4, 5],
            allow_nan=False)

    return image


def read_dv(path, sanity_check=False):
    """Read a video file with ``dv`` extension.

    Parameters
    ----------
    path : str
        Path of the file to read.
    sanity_check : bool
        Check if the array returned fits with apifish pipeline.

    Returns
    -------
    video : ndarray
        Video read.

    """
    # check path
    check_parameter(
        path=str,
        sanity_check=bool)

    # read video file
    video = mrc.imread(path)

    # check the output video
    # metadata can be read running 'tensor.mrc.info()'
    if sanity_check:
        check_array(
            video,
            dtype=[np.uint16, np.int16, np.int32, np.float32],
            allow_nan=False)

    return video


def read_array(path):
    """Read a numpy array with ``npy`` extension.

    Parameters
    ----------
    path : str
        Path of the array to read.

    Returns
    -------
    array : ndarray
        Array read.

    """
    # check path
    check_parameter(path=str)

    # read array file
    array = np.load(path)

    return array


def read_array_from_csv(path, dtype=None, delimiter=";", encoding="utf-8",
                        skiprows=0):
    """Read a numpy array saved in a ``csv`` file.

    Parameters
    ----------
    path : str
        Path of the ``csv`` file to read.
    dtype : type, optional
        Expected dtype to cast the array.
    delimiter : str, default=";"
        Delimiter used to separate columns.
    encoding : str, default="utf-8"
        Encoding to use.
    skiprows : int, default=0
        Skip the first `skiprows` lines of the file. Useful to skip the first
        rows of a csv with header.

    Returns
    -------
    array : ndarray
        Array read.

    """
    # check parameters
    check_parameter(
        path=str,
        dtype=(type, type(None)),
        delimiter=str,
        encoding=str)

    # read csv file
    array = np.loadtxt(
        path,
        delimiter=delimiter,
        encoding=encoding,
        skiprows=skiprows)

    # cast array dtype
    if dtype is not None:
        array = array.astype(dtype)

    return array


def read_dataframe_from_csv(path, delimiter=";", encoding="utf-8"):
    """Read a numpy array or a pandas object saved in a ``csv`` file.

    Parameters
    ----------
    path : str
        Path of the ``csv`` file to read.
    delimiter : str
        Delimiter used to separate columns.
    encoding : str
        Encoding to use.

    Returns
    -------
    df : pd.DataFrame
        Pandas object read.

    """
    # check parameters
    check_parameter(
        path=str,
        delimiter=str,
        encoding=str)

    # read csv file
    df = pd.read_csv(path, sep=delimiter, encoding=encoding)

    return df


def read_table_from_ecsv(path):
    """Read an astropy Table saved as an ``ecsv`` file.

    Parameters
    ----------
    path : str
        Path of the ``csv`` file to read.
    delimiter : str
        Delimiter used to separate columns.
    encoding : str
        Encoding to use.

    Returns
    -------
    df : pd.DataFrame
        Pandas object read.


    """
    # check parameters
    check_parameter(
        path=str,)

    # read ecsv file
    table = Table.read(path, format="ascii.ecsv")

    return table


def read_uncompressed(path, verbose=False):
    """Read a NpzFile object with ``npz`` extension.

    Parameters
    ----------
    path : str
        Path of the file to read.
    verbose : bool
        Return names of the different objects.

    Returns
    -------
    data : NpzFile object
        NpzFile read.

    """
    # check parameters
    check_parameter(
        path=str,
        verbose=bool)

    # read array file
    data = np.load(path)
    if verbose:
        print("Available keys: {0} \n".format(", ".join(data.files)))

    return data


def read_cell_extracted(path, verbose=False):
    """Read a NpzFile object with ``npz`` extension, previously written with
    :func:`apifish.stack.save_cell_extracted`.

    Parameters
    ----------
    path : str
        Path of the file to read.
    verbose : bool
        Return names of the different objects.

    Returns
    -------
    cell_results : Dict
        Dictionary including information about the cell (image, masks,
        coordinates arrays). Minimal information are:

        * `cell_id`: Unique id of the cell.
        * `bbox`: bounding box coordinates with the order (`min_y`, `min_x`,
          `max_y`, `max_x`).
        * `cell_coord`: boundary coordinates of the cell.
        * `cell_mask`: mask of the cell.

    """
    # read compressed file
    data = read_uncompressed(path, verbose)

    # store data in a dictionary
    cell_results = {}
    for key in data.files:
        cell_results[key] = data[key]

    return cell_results


# ### Write ###

def save_image(image, path, extension="tif"):
    """Save an image.

    The input image should have between 2 and 5 dimensions, with boolean,
    (unsigned) integer, or float.

    The dimensions should be in the following order: (round, channel, z, y, x).

    Parameters
    ----------
    image : np.ndarray
        Image to save.
    path : str
        Path of the saved image.
    extension : str
        Default extension to save the image (among ``png``, ``jpg``, ``jpeg``,
        ``tif`` or ``tiff``).

    Notes
    -----
    * If the image has more than 2 dimensions, ``tif`` and ``tiff`` extensions
      are required (``png`` extension does not handle 3-d images other than
      (M, N, 3) or (M, N, 4) shapes).
    * A 2-d boolean image can be saved in ``png``, ``jpg`` or ``jpeg`` (cast in
      np.uint8).
    * A multidimensional boolean image should be saved with
      :func:`apifish.stack.save_array` or as a boolean images with ``tif``/
      ``tiff`` extension.

    """
    # check image and parameters
    check_parameter(
        path=str,
        extension=str)
    check_array(
        image,
        dtype=[np.uint8, np.uint16, np.uint32, np.uint64,
               np.int8, np.int16, np.int32, np.int64,
               np.float16, np.float32, np.float64,
               bool],
        ndim=[2, 3, 4, 5],
        allow_nan=False)

    # check extension and build path
    if "/" in path:
        path_ = path.split("/")
        filename = path_[-1]

        if "." in filename:
            extension = filename.split(".")[-1]
        else:
            filename += ".{0}".format(extension)
        path_[-1] = filename
        path = "/".join(path_)
    else:
        if "." in path:
            extension = path.split(".")[-1]
        else:
            path += ".{0}".format(extension)
    if extension not in ["png", "jpg", "jpeg", "tif", "tiff"]:
        raise ValueError("{0} extension is not supported, please choose among "
                         "'png', 'jpg', 'jpeg', 'tif' or 'tiff'."
                         .format(extension))

    # warn about extension
    if (extension in ["png", "jpg", "jpeg"] and len(image.shape) > 2
            and image.dtype != bool):
        raise ValueError("Extension {0} is not fitted with multidimensional "
                         "images. Use 'tif' or 'tiff' extension instead."
                         .format(extension))
    if (extension in ["png", "jpg", "jpeg"] and len(image.shape) == 2
            and image.dtype != bool):
        warnings.warn("Extension {0} is not consistent with dtype. To prevent "
                      "'image' from being cast you should use 'tif' or 'tiff' "
                      "extension.".format(extension), UserWarning)
    if (extension in ["png", "jpg", "jpeg"] and len(image.shape) == 2
            and image.dtype == bool):
        warnings.warn("Extension {0} is not consistent with dtype. To prevent "
                      "'image' from being cast you should use "
                      "'apifish.stack.save_array' function instead."
                      .format(extension), UserWarning)
    if (extension in ["tif", "tiff"] and len(image.shape) == 2
            and image.dtype == bool):
        raise ValueError("Extension {0} is not fitted with boolean images. "
                         "Use 'png', 'jpg' or 'jpeg' extension instead."
                         .format(extension))
    if (extension in ["png", "jpg", "jpeg", "tif", "tiff"]
            and len(image.shape) > 2 and image.dtype == bool):
        raise ValueError("Extension {0} is not fitted with multidimensional "
                         "boolean images. Use 'apifish.stack.save_array' "
                         "function instead.".format(extension))

    # save image without warnings
    with warnings.catch_warnings():
        warnings.filterwarnings(action="ignore", category=UserWarning)
        io.imsave(path, image)


def save_array(array, path):
    """Save an array in a ``npy`` extension file.

    The input array should have between 2 and 5 dimensions, with boolean,
    (unsigned) integer, or float.

    Parameters
    ----------
    array : np.ndarray
        Array to save.
    path : str
        Path of the saved array.

    """
    # check array and path
    check_parameter(path=str)
    check_array(
        array,
        dtype=[np.uint8, np.uint16, np.uint32, np.uint64,
               np.int8, np.int16, np.int32, np.int64,
               np.float16, np.float32, np.float64,
               bool],
        ndim=[2, 3, 4, 5])

    # add extension if necessary
    if ".npy" not in path:
        path += ".npy"

    # save array
    np.save(path, array)


def save_data_to_csv(data, path, delimiter=";"):
    """Save a numpy array or a pandas object into a ``csv`` file.

    The input should be a pandas object (`Series` or `DataFrame`) or a
    numpy array with 2 dimensions and (unsigned) integer or float.

    Parameters
    ----------
    data : np.ndarray, pd.Series or pd.DataFrame
        Data to save.
    path : str
        Path of the saved ``csv`` file.
    delimiter : str
        Delimiter used to separate columns.

    """
    # check parameters
    check_parameter(
        data=(pd.DataFrame, pd.Series, np.ndarray),
        path=str,
        delimiter=str)

    # add extension if necessary
    if ".csv" not in path:
        path += ".csv"

    # save numpy ndarray in a csv file
    if not isinstance(data, (pd.DataFrame, pd.Series)):
        check_array(
            data,
            dtype=[np.uint8, np.uint16, np.uint32, np.uint64,
                   np.int8, np.int16, np.int32, np.int64,
                   np.float16, np.float32, np.float64],
            ndim=2)

        if data.dtype == np.float16:
            fmt = "%.4f"
        elif data.dtype == np.float32:
            fmt = "%.7f"
        elif data.dtype == np.float64:
            fmt = "%.16f"
        else:
            fmt = "%.1i"
        np.savetxt(path, data, fmt=fmt, delimiter=delimiter, encoding="utf-8")

    # save pandas object in a csv file
    elif isinstance(data, pd.Series):
        data = data.to_frame()
        data.to_csv(
            path,
            sep=delimiter,
            header=True,
            index=False,
            encoding="utf-8")
    else:
        data.to_csv(
            path,
            sep=delimiter,
            header=True,
            index=False,
            encoding="utf-8")


def save_table_to_ecsv(data, path):
    """Save an astropy table into an ``ecsv`` file.

    The input should be an astropy Table object

    Parameters
    ----------
    data : Table
        Data to save.
    path : str
        Path of the saved ``ecsv`` file.

    """
    # check parameters
    check_parameter(
        data=(Table),
        path=str,)

    data.write(
        path,
        format="ascii.ecsv",
        overwrite=True,
    )

def save_cell_extracted(cell_results, path):
    """Save cell-level results from :func:`apifish.stack.extract_cell` in a
    NpzFile object with ``npz`` extension.

    Parameters
    ----------
    cell_results : Dict
        Dictionary including information about the cell (image, masks,
        coordinates arrays). Minimal information are:

        * `cell_id`: Unique id of the cell.
        * `bbox`: bounding box coordinates with the order (`min_y`, `min_x`,
          `max_y`, `max_x`).
        * `cell_coord`: boundary coordinates of the cell.
        * `cell_mask`: mask of the cell.
    path : str
        Path of the saved array.

    """
    # check parameters
    check_parameter(
        cell_results=dict,
        path=str)

    # add extension if necessary
    if ".npz" not in path:
        path += ".npz"

    # save compressed file
    np.savez(path, **cell_results)



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for apifish.stack.quality module.
"""

import pytest

import numpy as np
import apifish.stack as stack


x_out_focus = np.array(
    [[[1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1]],

     [[1, 0, 0, 0, 0],
      [0, 1, 0, 0, 0],
      [0, 1, 0, 0, 0],
      [0, 1, 1, 1, 0],
      [0, 0, 0, 0, 0]],

     [[1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1]],

     [[2, 0, 0, 0, 0],
      [0, 1, 0, 0, 0],
      [0, 1, 0, 0, 0],
      [0, 1, 1, 1, 0],
      [0, 0, 0, 0, 0]],

     [[3, 0, 0, 0, 0],
      [0, 3, 0, 0, 3],
      [0, 3, 0, 0, 0],
      [0, 3, 3, 3, 0],
      [0, 0, 0, 0, 0]],

     [[1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1],
      [1, 1, 1, 1, 1]]],
    dtype=np.uint8)


@pytest.mark.parametrize("dtype", [
    np.uint8, np.uint16, np.float32, np.float64])
@pytest.mark.parametrize("neighborhood_size", [
    31, (5, 11), [11, 5]])
def test_compute_focus_format(dtype, neighborhood_size):
    # 2D input
    image = np.random.normal(loc=0.0, scale=1.0, size=10000)
    image = np.reshape(image, (100, 100)).astype(dtype)
    focus = stack.compute_focus(image, neighborhood_size=neighborhood_size)
    assert focus.dtype == np.float64
    assert focus.shape == image.shape
    assert focus.min() >= 1

    # 3D input
    image = np.random.normal(loc=0.0, scale=1.0, size=100000)
    image = np.reshape(image, (10, 100, 100)).astype(dtype)
    focus = stack.compute_focus(image, neighborhood_size=neighborhood_size)
    assert focus.dtype == np.float64
    assert focus.shape == image.shape
    assert focus.min() >= 1


def test_compute_focus_error():
    image = np.random.normal(loc=0.0, scale=1.0, size=10000)
    image = np.reshape(image, (100, 100))
    # error: sequence 'neighborhood_size' too short
    with pytest.raises(ValueError):
        _ = stack.compute_focus(image, neighborhood_size=[11])
    # error: sequence 'neighborhood_size' too long
    with pytest.raises(ValueError):
        _ = stack.compute_focus(image, neighborhood_size=(5, 11, 31))


def test_compute_focus():
    focus = stack.compute_focus(x_out_focus, neighborhood_size=31)
    average_focus = focus.mean(axis=(1, 2))
    assert round(average_focus[0], 5) == 1.
    assert round(average_focus[1], 5) > 1.
    assert round(average_focus[2], 5) == 1.
    assert round(average_focus[3], 5) > 1.
    assert round(average_focus[4], 5) > 1.
    assert round(average_focus[5], 5) == 1.



apifishapifishapifish# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for apifish.segmentation.cell_segmentation module.
"""


# TODO add test for apifish.segmentation.unet_distance_edge_double
# TODO add test for apifish.segmentation.apply_unet_distance_double
# TODO add test for apifish.segmentation.from_distance_to_instances
# TODO add test for apifish.segmentation.cell_watershed
# TODO add test for apifish.segmentation.get_watershed_relief
# TODO add test for apifish.segmentation.apply_watershed



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
The apifish.detection subpackage includes function to detect RNA spot in 2-d
and 3-d.
"""

from .spot_detection import detect_spots
from .spot_detection import local_maximum_detection
from .spot_detection import spots_thresholding
from .spot_detection import automated_threshold_setting
from .spot_detection import get_elbow_values

from .dense_decomposition import decompose_dense
from .dense_decomposition import get_dense_region
from .dense_decomposition import simulate_gaussian_mixture

from .spot_modeling import modelize_spot
from .spot_modeling import initialize_grid
from .spot_modeling import gaussian_2d
from .spot_modeling import gaussian_3d
from .spot_modeling import precompute_erf
from .spot_modeling import fit_subpixel

from .cluster_detection import detect_clusters

from .utils import convert_spot_coordinates
from .utils import get_object_radius_pixel
from .utils import get_object_radius_nm
from .utils import build_reference_spot
from .utils import compute_snr_spots
from .utils import get_breaking_point


_spots = [
    "detect_spots",
    "local_maximum_detection",
    "spots_thresholding",
    "automated_threshold_setting",
    "get_elbow_values"]

_dense = [
    "decompose_dense",
    "get_dense_region",
    "simulate_gaussian_mixture"]

_model = [
    "modelize_spot",
    "initialize_grid",
    "gaussian_2d",
    "gaussian_3d",
    "precompute_erf",
    "fit_subpixel"]

_clusters = [
    "detect_clusters"]

_utils = [
    "convert_spot_coordinates",
    "get_object_radius_pixel",
    "get_object_radius_nm",
    "build_reference_spot",
    "compute_snr_spots",
    "get_breaking_point"]

__all__ = _spots + _dense + _model + _clusters + _utils



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Segmentation models (utility functions).
"""

import tensorflow_addons as tfa

from tensorflow.python.keras.layers import Conv2D
from tensorflow.python.keras.layers import MaxPooling2D
from tensorflow.python.keras.layers import Add
from tensorflow.python.keras.layers import Activation
from tensorflow.python.keras.layers import UpSampling2D
from tensorflow.python.keras.engine.training import Model


# ### Convolution blocks ###

class SameConv(Model):

    def __init__(self,
                 filters,
                 kernel_size=(3, 3),
                 normalization=True,
                 activation="relu",
                 **kwargs):
        super(SameConv, self).__init__(**kwargs)

        # initialize parameters
        self.filters = filters
        self.kernel_size = kernel_size
        self.normalization = normalization
        self.activation = activation

        # define layers
        self.conv = Conv2D(
            filters=self.filters,
            kernel_size=self.kernel_size,
            padding='same')
        if self.normalization:
            self.norm = tfa.layers.InstanceNormalization()
        else:
            self.norm = None
        self.act = Activation(
            activation=self.activation)

    def call(self, inputs, training=False, mask=None):
        # compute layers
        x = self.conv(inputs)
        if self.normalization:
            x = self.norm(x)
        x = self.act(x)

        return x

    def get_config(self):
        config = {
            'filters': self.filters,
            'kernel_size': self.kernel_size,
            'normalization': self.normalization,
            'activation': self.activation}

        return config


class UpConv(Model):

    def __init__(self,
                 filters,
                 kernel_size=(3, 3),
                 normalization=True,
                 activation="relu",
                 interpolation="bilinear",
                 **kwargs):
        super(UpConv, self).__init__(**kwargs)

        # initialize parameters
        self.filters = filters
        self.kernel_size = kernel_size
        self.normalization = normalization
        self.activation = activation
        self.interpolation = interpolation

        # define layers
        self.resize = UpSampling2D(
            size=(2, 2),
            interpolation=self.interpolation)
        self.conv = SameConv(
            filters=self.filters,
            kernel_size=self.kernel_size,
            normalization=self.normalization,
            activation=self.activation)

    def call(self, inputs, training=False, mask=None):
        # resize and convolve image
        x = self.resize(inputs)
        x = self.conv(x)

        return x

    def get_config(self):
        config = {
            'filters': self.filters,
            'kernel_size': self.kernel_size,
            'normalization': self.normalization,
            'activation': self.activation,
            'interpolation': self.interpolation}

        return config


class DownBlock(Model):

    def __init__(self,
                 filters,
                 kernel_size=(3, 3),
                 normalization=True,
                 activation="relu",
                 pool_size=(2, 2),
                 **kwargs):
        super(DownBlock, self).__init__(**kwargs)

        # initialize parameters
        self.filters = filters
        self.kernel_size = kernel_size
        self.normalization = normalization
        self.activation = activation
        self.pool_size = pool_size

        # define layers
        self.conv_1 = SameConv(
            filters=self.filters,
            kernel_size=self.kernel_size,
            normalization=self.normalization,
            activation=self.activation)
        self.conv_2 = SameConv(
            filters=self.filters,
            kernel_size=self.kernel_size,
            normalization=self.normalization,
            activation=self.activation)
        self.residual = SameConv(
            filters=self.filters,
            kernel_size=(1, 1),
            normalization=self.normalization,
            activation=self.activation)
        self.add = Add()
        self.conv_3 = SameConv(
            filters=self.filters,
            kernel_size=self.kernel_size,
            normalization=self.normalization,
            activation=self.activation)
        self.conv_4 = SameConv(
            filters=self.filters,
            kernel_size=self.kernel_size,
            normalization=self.normalization,
            activation=self.activation)
        self.pool = MaxPooling2D(
            pool_size=self.pool_size)

    def call(self, inputs, training=False, mask=None):
        # two convolution layers
        x = self.conv_1(inputs)
        x = self.conv_2(x)

        # residual connection
        residual = self.residual(inputs)
        residual = self.add([x, residual])

        # two convolution layers
        x = self.conv_3(residual)
        x = self.conv_4(x)

        # pooling layer
        x = self.pool(x)

        return x, residual

    def get_config(self):
        config = {
            'filters': self.filters,
            'kernel_size': self.kernel_size,
            'normalization': self.normalization,
            'activation': self.activation,
            'pool_size': self.pool_size}

        return config


class UpBlock(Model):

    def __init__(self,
                 filters,
                 kernel_size=(3, 3),
                 normalization=True,
                 activation="relu",
                 interpolation="bilinear",
                 **kwargs):
        super(UpBlock, self).__init__(**kwargs)

        # initialize parameters
        self.filters = filters
        self.kernel_size = kernel_size
        self.normalization = normalization
        self.activation = activation
        self.interpolation = interpolation

        # define layers
        self.upsampling = UpConv(
            self.filters,
            kernel_size=self.kernel_size,
            normalization=self.normalization,
            activation=self.activation,
            interpolation=self.interpolation)
        self.add_1 = Add()
        self.conv_1 = SameConv(
            filters=self.filters,
            kernel_size=self.kernel_size,
            normalization=self.normalization,
            activation=self.activation)
        self.conv_2 = SameConv(
            filters=self.filters,
            kernel_size=self.kernel_size,
            normalization=self.normalization,
            activation=self.activation)
        self.add_2 = Add()
        self.conv_3 = SameConv(
            filters=self.filters,
            kernel_size=self.kernel_size,
            normalization=self.normalization,
            activation=self.activation)
        self.conv_4 = SameConv(
            filters=self.filters,
            kernel_size=self.kernel_size,
            normalization=self.normalization,
            activation=self.activation)

    def call(self, inputs, training=False, mask=None):
        # get inputs
        x, residual = inputs

        # upsampling
        x = self.upsampling(x)

        # residual connection
        residual = self.add_1([x, residual])

        # two convolution layers
        x = self.conv_1(residual)
        x = self.conv_2(x)

        # residual connection
        x = self.add_2([x, residual])

        # two convolution layers
        x = self.conv_3(x)
        x = self.conv_4(x)

        return x

    def get_config(self):
        config = {
            'filters': self.filters,
            'kernel_size': self.kernel_size,
            'normalization': self.normalization,
            'activation': self.activation,
            'interpolation': self.interpolation}

        return config


class Encoder(Model):

    def __init__(self,
                 **kwargs):
        super(Encoder, self).__init__(**kwargs)

        # define layers
        self.down_block_1 = DownBlock(filters=32)
        self.down_block_2 = DownBlock(filters=64)
        self.down_block_3 = DownBlock(filters=128)
        self.down_block_4 = DownBlock(filters=256)
        self.conv_1 = SameConv(filters=256)
        self.conv_2 = SameConv(filters=256)

    def call(self, inputs, training=False, mask=None):
        # (batch, H/2, W/2, 32), (batch, H, W, 32)
        x, residual_1 = self.down_block_1(inputs)

        # (batch, H/4, W/4, 64), (batch, H/2, W/2, 64)
        x, residual_2 = self.down_block_2(x)

        # (batch, H/8, W/8, 128), (batch, H/4, W/4, 128)
        x, residual_3 = self.down_block_3(x)

        # (batch, H/16, W/16, 256), (batch, H/8, W/8, 256)
        x, residual_4 = self.down_block_4(x)

        # bottom
        x = self.conv_1(x)
        x = self.conv_2(x)  # (batch, H/16, W/16, 256)

        return x, residual_1, residual_2, residual_3, residual_4

    def get_config(self):
        config = dict()

        return config


class Decoder(Model):

    def __init__(self,
                 **kwargs):
        super(Decoder, self).__init__(**kwargs)

        # define layers
        self.up_block_4 = UpBlock(filters=256)
        self.up_block_3 = UpBlock(filters=128)
        self.up_block_2 = UpBlock(filters=64)
        self.up_block_1 = UpBlock(filters=32)

    def call(self, inputs, training=False, mask=None):
        # get inputs
        x, residual_1, residual_2, residual_3, residual_4 = inputs

        # expansions
        x = self.up_block_4([x, residual_4])  # (batch, H/8, W/8, 256)
        x = self.up_block_3([x, residual_3])  # (batch, H/4, W/4, 128)
        x = self.up_block_2([x, residual_2])  # (batch, H/2, W/2, 64)
        x = self.up_block_1([x, residual_1])  # (batch, H, W, 32)

        return x

    def get_config(self):
        config = dict()

        return config


class EncoderDecoder(Model):

    def __init__(self,
                 **kwargs):
        super(EncoderDecoder, self).__init__(**kwargs)

        # initialize parameters
        self.encoder = Encoder()
        self.decoder = Decoder()

    def call(self, inputs, training=False, mask=None):
        # encode
        (x, residual_1, residual_2,
         residual_3, residual_4) = self.encoder(inputs)

        # decode
        x = self.decoder([x, residual_1, residual_2, residual_3, residual_4])

        return x

    def get_config(self):
        config = dict()

        return config



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Class and functions to segment cells.
"""

import apifish.stack as stack

from .utils import thresholding
from .postprocess import label_instances
from .postprocess import clean_segmentation

import numpy as np
from scipy import ndimage as ndi

import skimage
from sklearn.utils.fixes import parse_version
if parse_version(skimage.__version__) < parse_version("0.17.0"):
    from skimage.morphology import watershed
else:
    from skimage.segmentation import watershed


# ### Unet models ###

def unet_distance_edge_double():
    """Load a pretrained Unet model to predict foreground and a distance map
    to edge from nucleus and cell images.

    Returns
    -------
    model : ``tensorflow.keras.model`` object
        Pretrained Unet model.

    """
    # import  deep_learning subpackage
    import apifish.deep_learning as dl

    # load model
    model = dl.load_pretrained_model("double", "distance_edge")

    return model


def apply_unet_distance_double(model, nuc, cell, nuc_label, target_size=None,
                               test_time_augmentation=False):
    """Segment cell with a pretrained model to predict distance map and  use
    it with a watershed algorithm.

    Parameters
    ----------
    model : ``tensorflow.keras.model`` object
        Pretrained Unet model that predict distance to edges and cell surface.
    nuc : np.ndarray, np.uint
        Original nucleus image with shape (y, x).
    cell : np.ndarray, np.uint
        Original cell image to segment with shape (y, x).
    nuc_label : np.ndarray, np.int64
        Labelled nucleus image. Each nucleus is characterized by the same pixel
        value.
    target_size : int
        Resize image before segmentation. A squared image is resize to
        `target_size`. A rectangular image is resize such that its smaller
        dimension equals `target_size`.
    test_time_augmentation : bool
        Apply test time augmentation or not. The image is augmented 8 times
        and the final segmentation is the average result over these
        augmentations.

    Returns
    -------
    cell_label_pred : np.ndarray, np.int64
        Labelled cell image. Each cell is characterized by the same pixel
        value.

    """
    # check parameters
    stack.check_parameter(
        target_size=(int, type(None)),
        test_time_augmentation=bool)
    stack.check_array(nuc, ndim=2, dtype=[np.uint8, np.uint16])
    stack.check_array(cell, ndim=2, dtype=[np.uint8, np.uint16])
    stack.check_array(nuc_label, ndim=2, dtype=np.int64)

    # get original shape
    height, width = cell.shape

    # resize cell image if necessary
    if target_size is None:
        # keep the original shape
        nuc_to_process = nuc.copy()
        cell_to_process = cell.copy()
        nuc_label_to_process = nuc_label.copy()
        new_height, new_width = height, width

    else:
        # target size should be multiple of 16
        target_size -= 16

        # we resize the images below the target size
        ratio = target_size / min(height, width)
        new_height = int(np.round(height * ratio))
        new_width = int(np.round(width * ratio))
        new_shape = (new_height, new_width)
        nuc_to_process = stack.resize_image(nuc, new_shape, "bilinear")
        cell_to_process = stack.resize_image(cell, new_shape, "bilinear")
        nuc_label_to_process = nuc_label.copy()

    # get padding marge to make it multiple of 16
    marge_padding = stack.get_marge_padding(new_height, new_width, x=16)
    top, bottom = marge_padding[0]
    left, right = marge_padding[1]
    nuc_to_process = np.pad(
        nuc_to_process, pad_width=marge_padding, mode='symmetric')
    cell_to_process = np.pad(
        cell_to_process, pad_width=marge_padding, mode='symmetric')

    # standardize and cast cell image
    nuc_to_process = stack.compute_image_standardization(nuc_to_process)
    nuc_to_process = nuc_to_process.astype(np.float32)
    cell_to_process = stack.compute_image_standardization(cell_to_process)
    cell_to_process = cell_to_process.astype(np.float32)

    # augment images
    if test_time_augmentation:
        nuc_to_process = stack.augment_8_times(nuc_to_process)
        cell_to_process = stack.augment_8_times(cell_to_process)
        n_augmentations = 8
    else:
        nuc_to_process = [nuc_to_process]
        cell_to_process = [cell_to_process]
        n_augmentations = 1

    # loop over augmentations
    predictions_cell = []
    predictions_distance = []
    for i in range(n_augmentations):

        # get images
        nuc_to_process_ = nuc_to_process[i]
        nuc_to_process_ = nuc_to_process_[np.newaxis, :, :, np.newaxis]
        cell_to_process_ = cell_to_process[i]
        cell_to_process_ = cell_to_process_[np.newaxis, :, :, np.newaxis]

        # make predictions
        prediction = model.predict([nuc_to_process_, cell_to_process_])

        # remove padding
        if i in [0, 1, 2, 6]:
            prediction_cell = prediction[1][0, top:-bottom, left:-right, 0]
            prediction_distance = prediction[2][0, top:-bottom, left:-right, 0]
        else:
            prediction_cell = prediction[1][0, left:-right, top:-bottom, 0]
            prediction_distance = prediction[2][0, left:-right, top:-bottom, 0]

        # resize predictions back taking into account a potential deformation
        # from the image augmentation
        if target_size is not None:
            if i in [0, 1, 2, 6]:
                prediction_cell = stack.resize_image(
                    prediction_cell, (height, width), "bilinear")
                prediction_distance = stack.resize_image(
                    prediction_distance, (height, width), "bilinear")
            else:
                prediction_cell = stack.resize_image(
                    prediction_cell, (width, height), "bilinear")
                prediction_distance = stack.resize_image(
                    prediction_distance, (width, height), "bilinear")

        # store predictions
        predictions_cell.append(prediction_cell)
        predictions_distance.append(prediction_distance)

    # reversed image augmentation
    if test_time_augmentation:
        predictions_cell = stack.augment_8_times_reversed(predictions_cell)
        predictions_distance = stack.augment_8_times_reversed(
            predictions_distance)
        mean_prediction_cell = np.mean(predictions_cell, axis=0)
        mean_prediction_distance = np.mean(predictions_distance, axis=0)
    else:
        mean_prediction_cell = predictions_cell[0]
        mean_prediction_distance = predictions_distance[0]

    # inverse and format distance map
    min_ = mean_prediction_distance.min()
    max_ = max(1, mean_prediction_distance.max())
    mean_prediction_distance -= min_
    mean_prediction_distance /= max_
    mean_prediction_distance = 1 - mean_prediction_distance
    mean_prediction_distance = np.clip(mean_prediction_distance, 0, 1)
    mean_prediction_distance = stack.cast_img_uint16(mean_prediction_distance)

    # postprocess predictions
    _, cell_label_pred = from_distance_to_instances(
        label_x_nuc=nuc_label_to_process,
        label_2_cell=mean_prediction_cell,
        label_distance=mean_prediction_distance,
        compute_nuc_label=False)

    return cell_label_pred


def from_distance_to_instances(label_x_nuc, label_2_cell, label_distance,
                               nuc_3_classes=False, compute_nuc_label=False):
    """Extract instance labels from a distance map and a binary surface
    prediction with a watershed algorithm.

    Parameters
    ----------
    label_x_nuc : np.ndarray, np.float32
        Model prediction about the nucleus surface (and boundaries), with shape
        (y, x, 1) or (y, x, 3).
    label_2_cell : np.ndarray, np.float32
        Model prediction about cell surface, with shape (y, x, 1).
    label_distance : np.ndarray, np.uint16
        Model prediction about the distance to edges, with shape (y, x, 1).
    nuc_3_classes : bool
        Nucleus image input is an output from a 3-classes Unet.
    compute_nuc_label : bool
        Extract nucleus instance labels.

    Returns
    -------
    nuc_label : np.ndarray, np.int64
        Labelled nucleus image. Each nucleus is characterized by the same pixel
        value.
    cell_label : np.ndarray, np.int64
        Labelled cell image. Each cell is characterized by the same pixel
        value.

    """
    # check parameters
    stack.check_parameter(
        nuc_3_classes=bool,
        compute_nuc_label=bool)
    stack.check_array(label_x_nuc, ndim=2, dtype=[np.float32, np.int64])
    stack.check_array(label_2_cell, ndim=2, dtype=[np.float32])
    stack.check_array(label_distance, ndim=2, dtype=[np.uint16])

    # get nuclei labels
    if nuc_3_classes and compute_nuc_label:
        label_3_nuc = np.argmax(label_x_nuc, axis=-1)
        mask_nuc = label_3_nuc > 1
        nuc_label = label_instances(mask_nuc)
        nuc_label = nuc_label.astype(np.float64)
        nuc_label = stack.dilation_filter(
            nuc_label, kernel_shape="disk", kernel_size=1)
        nuc_label = nuc_label.astype(np.int64)
        mask_nuc = nuc_label > 0
    elif not nuc_3_classes and compute_nuc_label:
        mask_nuc = label_x_nuc > 0.5
        nuc_label = label_instances(mask_nuc)
    else:
        nuc_label = label_x_nuc.copy()
        mask_nuc = nuc_label > 0

    # get cells surfaces
    mask_cell = label_2_cell > 0.5
    mask_cell |= mask_nuc

    # apply watershed algorithm
    cell_label = watershed(label_distance, markers=nuc_label, mask=mask_cell)

    # cast labels in int64
    nuc_label = nuc_label.astype(np.int64)
    cell_label = cell_label.astype(np.int64)

    return nuc_label, cell_label


# ### Watershed ###

def cell_watershed(image, nuc_label, threshold, alpha=0.8):
    """Apply watershed algorithm to segment cell instances.

    In a watershed algorithm we consider cells as watershed to be flooded. The
    watershed relief is inversely proportional to both the pixel intensity and
    the closeness to nuclei. Pixels with a high intensity or close to labelled
    nuclei have a low watershed relief value. They will be flooded in priority.
    Flooding the watersheds allows to propagate nuclei labels through potential
    cytoplasm areas. The lines separating watershed are the final segmentation
    of the cells.

    Parameters
    ----------
    image : np.ndarray, np.uint
        Cells image with shape (y, x).
    nuc_label : np.ndarray, np.int64
        Result of the nuclei segmentation with shape (y, x) and nuclei
        instances labelled.
    threshold : int or float
        Threshold to discriminate cells surfaces from background.
    alpha : float or int
        Weight of the pixel intensity values to compute the watershed relief.

    Returns
    -------
    cell_label : np.ndarray, np.int64
        Segmentation of cells with shape (y, x).

    """
    # build relief
    relief = get_watershed_relief(image, nuc_label, alpha)

    # TODO improve cell mask methods
    # TODO add options for clean_segmentation
    # build cells mask
    if image.ndim == 3:
        image_2d = stack.maximum_projection(image)
    else:
        image_2d = image
    cell_mask = thresholding(image_2d, threshold)
    cell_mask[nuc_label > 0] = True
    cell_mask = clean_segmentation(
        cell_mask,
        small_object_size=5000,
        fill_holes=True)

    # segment cells
    cell_label = apply_watershed(relief, nuc_label, cell_mask)

    return cell_label


def get_watershed_relief(image, nuc_label, alpha):
    """Build a representation of cells as watershed.

    In a watershed algorithm we consider cells as watershed to be flooded. The
    watershed relief is inversely proportional to both the pixel intensity and
    the closeness to nuclei. Pixels with a high intensity or close to labelled
    nuclei have a low watershed relief value. They will be flooded in priority.
    Flooding the watersheds allows to propagate nuclei labels through potential
    cytoplasm areas. The lines separating watershed are the final segmentation
    of the cells.

    Parameters
    ----------
    image : np.ndarray, np.uint
        Cells image with shape (z, y, x) or (y, x).
    nuc_label : np.ndarray, np.int64
        Result of the nuclei segmentation with shape (y, x) and nuclei
        instances labelled.
    alpha : float or int
        Weight of the pixel intensity values to compute the relief.

    Returns
    -------
    watershed_relief : np.ndarray, np.uint16
        Watershed representation of cells with shape (y, x).

    """
    # check parameters
    stack.check_array(
        image,
        ndim=[2, 3],
        dtype=[np.uint8, np.uint16])
    stack.check_array(nuc_label, ndim=2, dtype=np.int64)
    stack.check_parameter(alpha=(int, float))

    # use pixel intensity of the cells image
    if alpha == 1:
        # if a 3-d image is provided we sum its pixel values
        image = stack.cast_img_float64(image)
        if image.ndim == 3:
            image = image.sum(axis=0)
        # rescale image
        image = stack.rescale(image)
        # build watershed relief
        watershed_relief = image.max() - image
        watershed_relief[nuc_label > 0] = 0
        watershed_relief = np.true_divide(
            watershed_relief,
            watershed_relief.max(),
            dtype=np.float64)
        watershed_relief = stack.cast_img_uint16(watershed_relief)

    # use distance from the nuclei
    elif alpha == 0:
        # build watershed relief
        nuc_mask = nuc_label > 0
        watershed_relief = ndi.distance_transform_edt(~nuc_mask)
        watershed_relief = np.true_divide(
            watershed_relief,
            watershed_relief.max(),
            dtype=np.float64)
        watershed_relief = stack.cast_img_uint16(watershed_relief)

    # use a combination of both previous methods
    elif 0 < alpha < 1:
        # if a 3-d image is provided we sum its pixel values
        image = stack.cast_img_float64(image)
        if image.ndim == 3:
            image = image.sum(axis=0)
        # rescale image
        image = stack.rescale(image)
        # build watershed relief
        relief_pixel = image.max() - image
        relief_pixel[nuc_label > 0] = 0
        relief_pixel = np.true_divide(
            relief_pixel,
            relief_pixel.max(),
            dtype=np.float64)
        nuc_mask = nuc_label > 0
        relief_distance = ndi.distance_transform_edt(~nuc_mask)
        relief_distance = np.true_divide(
            relief_distance,
            relief_distance.max(),
            dtype=np.float64)
        watershed_relief = alpha * relief_pixel + (1 - alpha) * relief_distance
        watershed_relief = stack.cast_img_uint16(watershed_relief)

    else:
        raise ValueError("Parameter 'alpha' is wrong. It must be comprised "
                         "between 0 and 1. Currently 'alpha' is {0}"
                         .format(alpha))

    return watershed_relief


def apply_watershed(watershed_relief, nuc_label, cell_mask):
    """Apply watershed algorithm to segment cell instances.

    In a watershed algorithm we consider cells as watershed to be flooded. The
    watershed relief is inversely proportional to both the pixel intensity and
    the closeness to nuclei. Pixels with a high intensity or close to labelled
    nuclei have a low watershed relief value. They will be flooded in priority.
    Flooding the watersheds allows to propagate nuclei labels through potential
    cytoplasm areas. The lines separating watershed are the final segmentation
    of the cells.

    Parameters
    ----------
    watershed_relief : np.ndarray, np.uint or np.int
        Watershed representation of cells with shape (y, x).
    nuc_label : np.ndarray, np.int64
        Result of the nuclei segmentation with shape (y, x) and nuclei
        instances labelled.
    cell_mask : np.ndarray, bool
        Binary image of cells surface with shape (y, x).

    Returns
    -------
    cell_label : np.ndarray, np.int64
        Segmentation of cells with shape (y, x).

    """
    # check parameters
    stack.check_array(
        watershed_relief,
        ndim=2,
        dtype=[np.uint8, np.uint16, np.int64])
    stack.check_array(nuc_label, ndim=2, dtype=np.int64)
    stack.check_array(cell_mask, ndim=2, dtype=bool)

    # segment cells
    cell_label = watershed(watershed_relief, markers=nuc_label, mask=cell_mask)
    cell_label = cell_label.astype(np.int64)

    return cell_label



apifishapifishapifish# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Utility functions for apifish.segmentation subpackage.
"""

import apifish.stack as stack

import numpy as np

from skimage.measure import regionprops


# TODO make functions compatible with different type of integers

# ### Thresholding method ###

def thresholding(image, threshold):
    """Segment a 2-d image to discriminate object from background applying a
    threshold.

    Parameters
    ----------
    image : np.ndarray, np.uint
        A 2-d image to segment with shape (y, x).
    threshold : int or float
        Pixel intensity threshold used to discriminate foreground from
        background.

    Returns
    -------
    image_segmented : np.ndarray, bool
        Binary 2-d image with shape (y, x).

    """
    # check parameters
    stack.check_array(image, ndim=2, dtype=[np.uint8, np.uint16])
    stack.check_parameter(threshold=(float, int))

    # discriminate nuclei from background, applying a threshold.
    image_segmented = image >= threshold

    return image_segmented


# ### Instances measures ###

def compute_mean_diameter(image_label):
    """Compute the averaged size of the segmented instances.

    For each instance, we compute the diameter of an object with an equivalent
    surface. Then, we average the diameters.

    Parameters
    ----------
    image_label : np.ndarray, np.int or np.uint
        Labelled image with shape (y, x).

    Returns
    -------
    mean_diameter : float
        Averaged size of the segmented instances.

    """
    # check parameters
    stack.check_array(
        image_label,
        ndim=2,
        dtype=[np.uint8, np.uint16, np.int64])

    # compute properties of the segmented instances
    props = regionprops(image_label)

    # get equivalent diameter and average it
    diameter = 0
    n = len(props)
    for prop in props:
        diameter += prop.equivalent_diameter
    if n > 0:
        mean_diameter = diameter / n
    else:
        mean_diameter = np.nan

    return mean_diameter


def compute_mean_convexity_ratio(image_label):
    """Compute the averaged convexity ratio of the segmented instances.

    For each instance, we compute the ratio between its area and the area of
    its convex hull. Then, we average the diameters.

    Parameters
    ----------
    image_label : np.ndarray, np.int or np.uint
        Labelled image with shape (y, x).

    Returns
    -------
    mean_convexity_ratio : float
        Averaged convexity ratio of the segmented instances.

    """
    # check parameters
    stack.check_array(
        image_label,
        ndim=2,
        dtype=[np.uint8, np.uint16, np.int64])

    # compute properties of the segmented instances
    props = regionprops(image_label)

    # get convexity ratio and average it
    n = len(props)
    convexity_ratio = 0
    for prop in props:
        convexity_ratio += prop.area / prop.convex_area
    if n > 0:
        mean_convexity_ratio = convexity_ratio / n
    else:
        mean_convexity_ratio = np.nan

    return mean_convexity_ratio


def compute_surface_ratio(image_label):
    """Compute the averaged surface ratio of the segmented instances.

    We compute the proportion of surface occupied by instances.

    Parameters
    ----------
    image_label : np.ndarray, np.int or np.uint
        Labelled image with shape (y, x).

    Returns
    -------
    surface_ratio : float
        Surface ratio of the segmented instances.

    """
    # check parameters
    stack.check_array(
        image_label,
        ndim=2,
        dtype=[np.uint8, np.uint16, np.int64])

    # compute surface ratio
    surface_instances = image_label > 0
    area_instances = surface_instances.sum()
    surface_ratio = area_instances / image_label.size

    return surface_ratio


def count_instances(image_label):
    """Count the number of instances annotated in the image.

    Parameters
    ----------
    image_label : np.ndarray, np.int or np.uint
        Labelled image with shape (y, x).

    Returns
    -------
    nb_instances : int
        Number of instances in the image.

    """
    # check parameters
    stack.check_array(
        image_label,
        ndim=2,
        dtype=[np.uint8, np.uint16, np.int64])

    indices = set(image_label.ravel())
    if 0 in indices:
        nb_instances = len(indices) - 1
    else:
        nb_instances = len(indices)

    return nb_instances



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Utility functions for apifish.multistack subpackage.
"""

import os
import re
import copy

import apifish.stack as stack


# ### Recipe management (sanity checks, fitting) ###

def check_recipe(recipe, data_directory=None):
    """Check and validate a recipe.

    Checking a recipe consists in validating its filename pattern and the
    content of the dictionary.

    Parameters
    ----------
    recipe : dict
        Map the images according to their field of view, their round,
        their channel and their spatial dimensions. Can only contain the keys
        `pattern`, `fov`, `r`, `c`, `z`, `ext` or `opt`.
    data_directory : str
        Path of the directory with the files describes in the recipe. If it is
        provided, the function check that the files exist.

    Returns
    -------
    _ : bool
        Assert if the recipe is well formatted.

    """
    # check parameters
    stack.check_parameter(
        recipe=dict,
        data_directory=(str, type(None)))

    # check the filename pattern
    if "pattern" not in recipe:
        raise KeyError("A recipe should have a filename pattern "
                       "('pattern' keyword).")
    recipe_pattern = recipe["pattern"]
    if not isinstance(recipe_pattern, str):
        raise TypeError("'pattern' should be a string, not a {0}."
                        .format(type(recipe_pattern)))

    # count the different dimensions to combinate in the recipe (among
    # 'fov', 'r', 'c' and 'z')
    dimensions = re.findall("fov|r|c|z", recipe_pattern)

    # each dimension can only appear once in the filename pattern
    if len(dimensions) != len(set(dimensions)):
        raise ValueError("The pattern used in recipe is wrong, a dimension "
                         "appears several times: {0}".format(recipe_pattern))

    # check keys and values of the recipe
    for key, value in recipe.items():
        if key not in ['fov', 'r', 'c', 'z', 'ext', 'opt', 'pattern']:
            raise KeyError("The recipe can only contain the keys 'fov', 'r', "
                           "'c', 'z', 'ext', 'opt' or 'pattern'. Not '{0}'."
                           .format(key))
        if not isinstance(value, (list, str)):
            raise TypeError("A recipe can only contain lists or strings, "
                            "not {0}.".format(type(value)))

    # check that requested files exist
    if data_directory is not None:
        if not os.path.isdir(data_directory):
            raise NotADirectoryError("Directory does not exist: {0}"
                                     .format(data_directory))
        recipe = fit_recipe(recipe)
        nb_r, nb_c, nb_z = get_nb_element_per_dimension(recipe)
        nb_fov = count_nb_fov(recipe)
        for fov in range(nb_fov):
            for r in range(nb_r):
                for c in range(nb_c):
                    for z in range(nb_z):
                        path = get_path_from_recipe(
                            recipe,
                            data_directory,
                            fov=fov,
                            r=r,
                            c=c,
                            z=z)
                        if not os.path.isfile(path):
                            raise FileNotFoundError("File does not exist: {0}"
                                                    .format(path))

    return True


def fit_recipe(recipe):
    """Fit a recipe.

    Fitting a recipe consists in wrapping every values of `fov`, `r`, `c` and
    `z` in a list (an empty one if necessary). Values for `ext` and `opt` are
    also initialized.

    Parameters
    ----------
    recipe : dict
        Map the images according to their field of view, their round,
        their channel and their spatial dimensions. Can only contain the keys
        `pattern`, `fov`, `r`, `c`, `z`, `ext` or `opt`.

    Returns
    -------
    new_recipe : dict
        Map the images according to their field of view, their round,
        their channel and their spatial dimensions. Contain the keys
        `pattern`, `fov`, `r`, `c`, `z`, `ext` and `opt`, initialized if
        necessary.

    """
    # check parameters
    stack.check_parameter(recipe=dict)

    # initialize recipe
    new_recipe = copy.deepcopy(recipe)

    # initialize and fit the dimensions 'fov', 'r', 'c' and 'z'
    for key in ['fov', 'r', 'c', 'z']:
        if key not in new_recipe:
            new_recipe[key] = [None]
        value = new_recipe[key]
        if isinstance(value, str):
            new_recipe[key] = [value]

    # initialize the dimensions 'ext', 'opt'
    for key in ['ext', 'opt']:
        if key not in new_recipe:
            new_recipe[key] = ""

    return new_recipe


def _is_recipe_fitted(recipe):
    """Check if a recipe is ready to be used.

    Fitting a recipe consists in wrapping every values of `fov`, `r`, `c` and
    `z` in a list (an empty one if necessary). Values for `ext` and `opt` are
    also initialized.

    Parameters
    ----------
    recipe : dict
        Map the images according to their field of view, their round,
        their channel and their spatial dimensions. Can only contain the keys
        `pattern`, `fov`, `r`, `c`, `z`, `ext` or `opt`.

    Returns
    -------
    _ : bool
        Indicates if the recipe is fitted or not

    """
    # all keys should be initialized in the new recipe, with a list or a string
    for key in ['fov', 'r', 'c', 'z']:
        if key not in recipe or not isinstance(recipe[key], list):
            return False
    for key in ['ext', 'opt']:
        if key not in recipe or not isinstance(recipe[key], str):
            return False
    if 'pattern' not in recipe or not isinstance(recipe['pattern'], str):
        return False

    return True


def get_path_from_recipe(recipe, input_folder, fov=0, r=0, c=0, z=0):
    """Build the path of a file from a recipe and the indices of specific
    elements.

    Parameters
    ----------
    recipe : dict
        Map the images according to their field of view, their round,
        their channel and their spatial dimensions. Only contain the keys
        `pattern`, `fov`, `r`, `c`, `z`, `ext` or `opt`.
    input_folder : str
        Path of the folder containing the images.
    fov : int
        Index of the `fov` element in the recipe to use in the filename.
    r : int
        Index of the `r` element in the recipe to use in the filename.
    c : int
        Index of the `c` element in the recipe to use in the filename.
    z : int
        Index of the `z` element in the recipe to use in the filename.

    Returns
    -------
    path : str
        Path of the file to load.

    """
    # check parameters
    stack.check_parameter(
        recipe=dict,
        input_folder=str,
        fov=int,
        r=int,
        c=int,
        z=int)

    # check if the recipe is fitted
    if not _is_recipe_fitted(recipe):
        recipe = fit_recipe(recipe)

    # build a map of the elements' indices
    map_element_index = {"fov": fov, "r": r, "c": c, "z": z}

    # get filename pattern and decompose it
    recipe_pattern = recipe["pattern"]
    path_elements = re.findall("fov|r|c|z|ext|opt", recipe_pattern)
    path_separators = re.split("fov|r|c|z|ext|opt", recipe_pattern)

    # get filename recombining elements of the recipe
    filename = path_separators[0]  # usually an empty string
    for (element_name, separator) in zip(path_elements, path_separators[1:]):

        # if we need an element from a list of elements of the same dimension
        # (eg. to pick a specific channel 'c' among a list of channels)
        if element_name in map_element_index:
            element_index = map_element_index[element_name]
            element = recipe[element_name][element_index]

        # if this element is unique for all the recipe (eg. 'fov')
        else:
            element = recipe[element_name]

        # the filename is built ensuring the order of apparition of the
        # different morphemes and their separators
        filename += element
        filename += separator

    # get path
    path = os.path.join(input_folder, filename)

    return path


def get_nb_element_per_dimension(recipe):
    """Count the number of element to stack for each dimension (`r`, `c`
    and `z`).

    Parameters
    ----------
    recipe : dict
        Map the images according to their field of view, their round,
        their channel and their spatial dimensions. Only contain the keys
        `fov`, `r`, `c`, `z`, `ext` or `opt`.

    Returns
    -------
    nb_r : int
        Number of rounds to be stacked.
    nb_c : int
        Number of channels to be stacked.
    nb_z : int
        Number of z layers to be stacked.

    """
    # check parameters
    stack.check_parameter(recipe=dict)

    # check if the recipe is fitted
    if not _is_recipe_fitted(recipe):
        recipe = fit_recipe(recipe)

    return len(recipe["r"]), len(recipe["c"]), len(recipe["z"])


def count_nb_fov(recipe):
    """Count the number of different fields of view that can be defined from
    the recipe.

    Parameters
    ----------
    recipe : dict
        Map the images according to their field of view, their round,
        their channel and their spatial dimensions. Can only contain the keys
        `pattern`, `fov`, `r`, `c`, `z`, `ext` or `opt`.

    Returns
    -------
    nb_fov : int
        Number of different fields of view in the recipe.

    """
    # check parameters
    stack.check_parameter(recipe=dict)

    # check if the recipe is fitted
    if not _is_recipe_fitted(recipe):
        recipe = fit_recipe(recipe)

    # a good recipe should have a list in the 'fov' key
    if not isinstance(recipe["fov"], list):
        raise TypeError("'fov' should be a List or a str, not {0}"
                        .format(type(recipe["fov"])))
    else:
        return len(recipe["fov"])


def check_datamap(data_map):
    """Check and validate a data map.

    Checking a data map consists in validating the recipe-folder pairs.

    Parameters
    ----------
    data_map : List[tuple]
        Map between input directories and recipes.

    Returns
    -------
    _ : bool
        Assert if the data map is well formatted.

    """
    stack.check_parameter(data_map=list)
    for pair in data_map:
        if not isinstance(pair, (tuple, list)):
            raise TypeError("A data map is a list with tuples or lists. "
                            "Not {0}".format(type(pair)))
        if len(pair) != 2:
            raise ValueError("Elements of a data map are tuples or lists that "
                             "map a recipe (dict) to an input directory "
                             "(string). Here {0} elements are given {1}"
                             .format(len(pair), pair))
        (recipe, input_folder) = pair
        if not isinstance(input_folder, str):
            raise TypeError("A data map map a recipe (dict) to an input "
                            "directory (string). Not ({0}, {1})"
                            .format(type(recipe), type(input_folder)))
        check_recipe(recipe, data_directory=input_folder)

    return True



apifishapifishapifish# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
The apifish.segmentation subpackage includes functions to segment or label
nuclei and cells.
"""

from .cell_segmentation import unet_distance_edge_double
from .cell_segmentation import apply_unet_distance_double
from .cell_segmentation import from_distance_to_instances
from .cell_segmentation import cell_watershed
from .cell_segmentation import get_watershed_relief
from .cell_segmentation import apply_watershed

from .nuc_segmentation import unet_3_classes_nuc
from .nuc_segmentation import apply_unet_3_classes
from .nuc_segmentation import from_3_classes_to_instances
from .nuc_segmentation import remove_segmented_nuc

from .postprocess import label_instances
from .postprocess import merge_labels
from .postprocess import clean_segmentation
from .postprocess import remove_disjoint

from .utils import thresholding
from .utils import compute_mean_diameter
from .utils import compute_mean_convexity_ratio
from .utils import compute_surface_ratio
from .utils import count_instances


_cell = [
    "unet_distance_edge_double",
    "apply_unet_distance_double",
    "from_distance_to_instances",
    "cell_watershed",
    "get_watershed_relief",
    "apply_watershed"]

_nuc = [
    "unet_3_classes_nuc",
    "apply_unet_3_classes",
    "from_3_classes_to_instances",
    "remove_segmented_nuc"]

_postprocess = [
    "label_instances",
    "merge_labels",
    "clean_segmentation",
    "remove_disjoint"]

_utils = [
    "thresholding",
    "compute_mean_diameter",
    "compute_mean_convexity_ratio",
    "compute_surface_ratio",
    "count_instances"]

__all__ = _cell + _nuc + _postprocess + _utils



apifishapifishapifish# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Unitary tests for apifish.multistack.preprocess module.
"""

import os
import pytest
import tempfile

import numpy as np

import apifish.stack as stack
import apifish.multistack as multistack

from numpy.testing import assert_array_equal


# ### Test stack building ###

def test_build_stacks_from_recipe():
    # build a temporary directory and save tensors inside
    with tempfile.TemporaryDirectory() as tmp_dir:
        # field of view 1
        test_nuc = np.zeros((8, 8, 8), dtype=np.uint8)
        test_cyt = np.zeros((8, 8, 8), dtype=np.uint8)
        test_rna = np.zeros((8, 8, 8), dtype=np.uint8)
        path_nuc = os.path.join(tmp_dir, "test_nuc_1.tif")
        path_cyt = os.path.join(tmp_dir, "test_cyt_1.tif")
        path_rna = os.path.join(tmp_dir, "test_rna_1.tif")
        stack.save_image(test_nuc, path_nuc)
        stack.save_image(test_cyt, path_cyt)
        stack.save_image(test_rna, path_rna)

        # field of view 2
        test_nuc = np.zeros((5, 5, 5), dtype=np.uint16)
        test_cyt = np.zeros((5, 5, 5), dtype=np.uint16)
        test_rna = np.zeros((5, 5, 5), dtype=np.uint16)
        path_nuc = os.path.join(tmp_dir, "test_nuc_2.tif")
        path_cyt = os.path.join(tmp_dir, "test_cyt_2.tif")
        path_rna = os.path.join(tmp_dir, "test_rna_2.tif")
        stack.save_image(test_nuc, path_nuc)
        stack.save_image(test_cyt, path_cyt)
        stack.save_image(test_rna, path_rna)

        # define recipe to read tensors
        recipe_1 = {"fov": ["1", "2"],
                    "c": ["nuc", "cyt", "rna"],
                    "opt": "test",
                    "ext": "tif",
                    "pattern": "opt_c_fov.ext"}

        # build tensor without prior information
        tensor = multistack.build_stack(recipe_1, input_folder=tmp_dir)
        expected_tensor = np.zeros((1, 3, 8, 8, 8), dtype=np.uint8)
        assert_array_equal(tensor, expected_tensor)
        assert tensor.dtype == np.uint8

        # build tensor with prior information
        tensor = multistack.build_stack(recipe_1,
                                        input_folder=tmp_dir,
                                        input_dimension=3)
        expected_tensor = np.zeros((1, 3, 8, 8, 8), dtype=np.uint8)
        assert_array_equal(tensor, expected_tensor)
        assert tensor.dtype == np.uint8

        # build tensors with different fields of view
        tensor = multistack.build_stack(recipe_1,
                                        input_folder=tmp_dir,
                                        input_dimension=3,
                                        i_fov=0)
        expected_tensor = np.zeros((1, 3, 8, 8, 8), dtype=np.uint8)
        assert_array_equal(tensor, expected_tensor)
        assert tensor.dtype == np.uint8
        tensor = multistack.build_stack(recipe_1,
                                        input_folder=tmp_dir,
                                        input_dimension=3,
                                        i_fov=1)
        expected_tensor = np.zeros((1, 3, 5, 5, 5), dtype=np.uint16)
        assert_array_equal(tensor, expected_tensor)
        assert tensor.dtype == np.uint16

        # wrong recipe
        recipe_wrong = {"fov": "test",
                        "c": ["nuc", "cyt", "rna"],
                        "ext": "tif",
                        "pattern": "fov_c.ext"}
        with pytest.raises(FileNotFoundError):
            multistack.build_stack(recipe_wrong,
                                   input_folder=tmp_dir,
                                   input_dimension=3)

        # wrong path
        with pytest.raises(FileNotFoundError):
            multistack.build_stack(recipe_1,
                                   input_folder="/foo/bar",
                                   input_dimension=3)


def test_build_stacks_from_datamap():
    # build a temporary directory and save tensors inside
    with tempfile.TemporaryDirectory() as tmp_dir:
        # field of view 1
        test_nuc = np.zeros((8, 8, 8), dtype=np.uint8)
        test_cyt = np.zeros((8, 8, 8), dtype=np.uint8)
        test_rna = np.zeros((8, 8, 8), dtype=np.uint8)
        path_nuc = os.path.join(tmp_dir, "test_nuc_1.tif")
        path_cyt = os.path.join(tmp_dir, "test_cyt_1.tif")
        path_rna = os.path.join(tmp_dir, "test_rna_1.tif")
        stack.save_image(test_nuc, path_nuc)
        stack.save_image(test_cyt, path_cyt)
        stack.save_image(test_rna, path_rna)

        # field of view 2
        test_nuc = np.zeros((5, 5, 5), dtype=np.uint16)
        test_cyt = np.zeros((5, 5, 5), dtype=np.uint16)
        test_rna = np.zeros((5, 5, 5), dtype=np.uint16)
        path_nuc = os.path.join(tmp_dir, "test_nuc_2.tif")
        path_cyt = os.path.join(tmp_dir, "test_cyt_2.tif")
        path_rna = os.path.join(tmp_dir, "test_rna_2.tif")
        stack.save_image(test_nuc, path_nuc)
        stack.save_image(test_cyt, path_cyt)
        stack.save_image(test_rna, path_rna)

        # define datamap to read  tensors
        recipe_1 = {"fov": ["1", "2"],
                    "c": ["nuc", "cyt", "rna"],
                    "opt": "test",
                    "ext": "tif",
                    "pattern": "opt_c_fov.ext"}
        recipe_2 = {"fov": "2",
                    "c": ["nuc", "cyt", "rna"],
                    "opt": "test",
                    "ext": "tif",
                    "pattern": "opt_c_fov.ext"}
        data_map = [(recipe_1, tmp_dir), (recipe_2, tmp_dir)]

        # build stacks from generator
        generator = multistack.build_stacks(data_map, input_dimension=3)
        expected_tensors = [np.zeros((1, 3, 8, 8, 8), dtype=np.uint8),
                            np.zeros((1, 3, 5, 5, 5), dtype=np.uint16),
                            np.zeros((1, 3, 5, 5, 5), dtype=np.uint16)]
        for i, tensor in enumerate(generator):
            expected_tensor = expected_tensors[i]
            assert_array_equal(tensor, expected_tensor)
            assert tensor.dtype == expected_tensor.dtype

        # build stacks from generator with metadata
        generator = multistack.build_stacks(data_map,
                                            input_dimension=3,
                                            return_origin=True)
        expected_tensors = [np.zeros((1, 3, 8, 8, 8), dtype=np.uint8),
                            np.zeros((1, 3, 5, 5, 5), dtype=np.uint16),
                            np.zeros((1, 3, 5, 5, 5), dtype=np.uint16)]
        expected_recipes = [recipe_1, recipe_1, recipe_2]
        expected_i_fov = [0, 1, 0]
        for i, (tensor, input_folder, recipe, i_fov) in enumerate(generator):
            expected_tensor = expected_tensors[i]
            assert_array_equal(tensor, expected_tensor)
            assert tensor.dtype == expected_tensor.dtype
            assert input_folder == tmp_dir
            assert recipe == expected_recipes[i]
            assert i_fov == expected_i_fov[i]

        # wrong datamap
        data_map = [(recipe_1, 3), (recipe_2, tmp_dir)]
        generator = multistack.build_stacks(data_map, input_dimension=3)
        with pytest.raises(TypeError):
            next(generator)
        data_map = [(recipe_1, "foo/bar"), (recipe_2, tmp_dir)]
        generator = multistack.build_stacks(data_map, input_dimension=3)
        with pytest.raises(NotADirectoryError):
            next(generator)


def test_build_stack_from_path():
    # build a temporary directory and save tensors inside
    with tempfile.TemporaryDirectory() as tmp_dir:
        # field of view
        test_nuc = np.zeros((8, 8, 8), dtype=np.uint8)
        test_cyt = np.zeros((8, 8, 8), dtype=np.uint8)
        test_rna = np.zeros((8, 8, 8), dtype=np.uint8)
        path_nuc = os.path.join(tmp_dir, "test_nuc.tif")
        path_cyt = os.path.join(tmp_dir, "test_cyt.tif")
        path_rna = os.path.join(tmp_dir, "test_rna.tif")
        stack.save_image(test_nuc, path_nuc)
        stack.save_image(test_cyt, path_cyt)
        stack.save_image(test_rna, path_rna)

        # build tensor from paths
        paths = [path_nuc, path_cyt, path_rna]
        tensor = multistack.build_stack_no_recipe(paths, input_dimension=3)
        expected_tensor = np.zeros((1, 3, 8, 8, 8), dtype=np.uint8)
        assert_array_equal(tensor, expected_tensor)
        assert tensor.dtype == np.uint8

        # wrong paths
        paths = [path_nuc, path_cyt, "/foo/bar/test_rna.tif"]
        with pytest.raises(FileNotFoundError):
            multistack.build_stack_no_recipe(paths, input_dimension=3)



# -*- coding: utf-8 -*-
# Author: Arthur Imbert <arthur.imbert.pro@gmail.com>
# License: BSD 3 clause

"""
Functions used to format and clean any input loaded in apifish.
"""

import warnings

import numpy as np

from .utils import check_array
from .utils import check_parameter
from .utils import check_range_value

import skimage
from skimage import img_as_ubyte
from skimage import img_as_float32
from skimage import img_as_float64
from skimage import img_as_uint
from skimage.exposure import rescale_intensity
from skimage.transform import resize
from sklearn.utils.fixes import parse_version


# TODO replace 'tensor' by 'image'

# ### Image normalization ###

def compute_image_standardization(image):
    """Normalize image by computing its z score.

    Parameters
    ----------
    image : np.ndarray
        Image to normalize with shape (y, x).

    Returns
    -------
    normalized_image : np.ndarray
        Normalized image with shape (y, x).

    """
    # check parameters
    check_array(image, ndim=2, dtype=[np.uint8, np.uint16, np.float32])

    # check image is in 2D
    if len(image.shape) != 2:
        raise ValueError("'image' should be a 2-d array. Not {0}-d array"
                         .format(len(image.shape)))

    # compute mean and standard deviation
    m = np.mean(image)
    adjusted_stddev = max(np.std(image), 1.0 / np.sqrt(image.size))

    # normalize image
    normalized_image = (image - m) / adjusted_stddev

    return normalized_image


def rescale(tensor, channel_to_stretch=None, stretching_percentile=99.9):
    """Rescale tensor values up to its dtype range (unsigned/signed integers)
    or between 0 and 1 (float).

    Each round and each channel is rescaled independently. Tensor has between
    2 to 5 dimensions, in the following order: (round, channel, z, y, x).

    By default, we rescale the tensor intensity range to its dtype range (or
    between 0 and 1 for float tensor). We can improve the contrast by
    stretching a smaller range of pixel intensity: between the minimum value
    of a channel and percentile value of the channel (cf.
    ``stretching_percentile``).

    To be consistent with skimage, 64-bit (unsigned) integer images are not
    supported.

    Parameters
    ----------
    tensor : np.ndarray
        Tensor to rescale.
    channel_to_stretch : int, List[int] or Tuple[int]
        Channel to stretch. If None, minimum and maximum of each channel are
        used as the intensity range to rescale.
    stretching_percentile : float or int
        Percentile to determine the maximum intensity value used to rescale
        the image. If 1, the maximum pixel intensity is used to rescale the
        image.

    Returns
    -------
    tensor : np.ndarray
        Tensor rescaled.

    """
    # check parameters
    check_parameter(
        tensor=np.ndarray,
        channel_to_stretch=(int, list, tuple, type(None)),
        stretching_percentile=(int, float))
    check_array(
        tensor,
        ndim=[2, 3, 4, 5],
        dtype=[np.uint8, np.uint16, np.uint32,
               np.int8, np.int16, np.int32,
               np.float32, np.float64])
    check_range_value(tensor, min_=0)

    # enlist 'channel_to_stretch' if necessary
    if channel_to_stretch is None:
        channel_to_stretch = []
    elif isinstance(channel_to_stretch, int):
        channel_to_stretch = [channel_to_stretch]

    # wrap tensor in 5-d if necessary
    tensor_5d, original_ndim = _wrap_5d(tensor)

    # rescale
    tensor_5d = _rescale_5d(
        tensor_5d,
        channel_to_stretch=channel_to_stretch,
        stretching_percentile=stretching_percentile)

    # rebuild the original tensor shape
    tensor = _unwrap_5d(tensor_5d, original_ndim)

    return tensor


def _wrap_5d(tensor):
    """Increases the number of dimensions of a tensor up to 5.

    Parameters
    ----------
    tensor : np.ndarray
        Tensor to wrap.

    Returns
    -------
    tensor_5d : np.ndarray
        Tensor with shape (round, channel, z, y, x).
    original_ndim : int
        Original number of dimensions.

    """
    # wrap tensor in 5-d if necessary
    original_ndim = tensor.ndim
    if original_ndim == 2:
        tensor_5d = tensor[np.newaxis, np.newaxis, np.newaxis, ...]
    elif original_ndim == 3:
        tensor_5d = tensor[np.newaxis, np.newaxis, ...]
    elif original_ndim == 4:
        tensor_5d = tensor[np.newaxis, ...]
    else:
        tensor_5d = tensor

    return tensor_5d, original_ndim


def _unwrap_5d(tensor_5d, original_ndim):
    """Remove useless dimensions from a 5-d tensor.

    Parameters
    ----------
    tensor_5d : np.ndarray
        Tensor with shape (round, channel, z, y, x).
    original_ndim : int
        Original number of dimensions.

    Returns
    -------
    tensor : np.ndarray
        Unwrapped tensor.

    """
    # rebuild the original tensor shape
    if original_ndim == 2:
        tensor = tensor_5d[0, 0, 0, :, :]
    elif original_ndim == 3:
        tensor = tensor_5d[0, 0, :, :, :]
    elif original_ndim == 4:
        tensor = tensor_5d[0, :, :, :, :]
    else:
        tensor = tensor_5d

    return tensor


def _rescale_5d(tensor, channel_to_stretch, stretching_percentile):
    """Rescale tensor values up to its dtype range (unsigned/signed integers)
    or between 0 and 1 (float).

    Each round and each channel is rescaled independently. Tensor has between
    2 to 5 dimensions, in the following order: (round, channel, z, y, x).

    By default, we rescale the tensor intensity range to its dtype range (or
    between 0 and 1 for float tensor). We can improve the contrast by
    stretching a smaller range of pixel intensity: between the minimum value
    of a channel and percentile value of the channel (cf.
    'stretching_percentile').

    Parameters
    ----------
    tensor : np.ndarray
        Tensor to rescale.
    channel_to_stretch : int, List[int] or Tuple[int]
        Channel to stretch. If None, minimum and maximum of each channel are
        used as the intensity range to rescale.
    stretching_percentile : float
        Percentile to determine the maximum intensity value used to rescale
        the image. If 1, the maximum pixel intensity is used to rescale the
        image.

    Returns
    -------
    tensor : np.ndarray
        Tensor rescaled.

    """
    # target intensity range
    target_range = 'dtype'
    dtype = tensor.dtype
    if dtype in [np.float16, np.float32, np.float64]:
        target_range = (0, 1)

    # rescale each round independently
    rounds = []
    for r in range(tensor.shape[0]):

        # rescale each channel independently
        channels = []
        for c in range(tensor.shape[1]):

            # get channel
            channel = tensor[r, c, :, :, :]

            # rescale channel
            if c in channel_to_stretch:
                pa, pb = np.percentile(channel, (0, stretching_percentile))
                channel_rescaled = rescale_intensity(
                    channel,
                    in_range=(pa, pb),
                    out_range=target_range)
            else:
                channel_rescaled = rescale_intensity(
                    channel,
                    out_range=target_range)
            channels.append(channel_rescaled)

        # stack channels
        tensor_4d = np.stack(channels, axis=0)
        rounds.append(tensor_4d)

    # stack rounds
    tensor_5d = np.stack(rounds, axis=0)

    # cast results
    tensor_5d = tensor_5d.astype(dtype)

    return tensor_5d


def cast_img_uint8(tensor):
    """Cast the image in np.uint8 and scale values between 0 and 255.

    Negative values are not allowed as the skimage method ``img_as_ubyte``
    would clip them to 0. Positives values are scaled between 0 and 255,
    excepted if they fit directly in 8 bit (in this case values are not
    modified).

    Parameters
    ----------
    tensor : np.ndarray
        Image to cast.

    Returns
    -------
    tensor : np.ndarray, np.uint8
        Image cast.

    """
    # check tensor dtype
    check_array(
        tensor,
        ndim=[2, 3, 4, 5],
        dtype=[np.uint8, np.uint16, np.uint32, np.uint64,
               np.int8, np.int16, np.int32, np.int64,
               np.float32, np.float64])
    if tensor.dtype in [np.float32, np.float64]:
        check_range_value(tensor, min_=0, max_=1)
    elif tensor.dtype in [np.int8, np.int16, np.int32, np.int64]:
        check_range_value(tensor, min_=0)

    if tensor.dtype == np.uint8:
        return tensor

    if (tensor.dtype in [np.uint16, np.uint32, np.uint64,
                         np.int16, np.int32, np.int64]
            and tensor.max() <= 255):
        raise ValueError("Tensor values are between {0} and {1}. It fits in 8 "
                         "bits and won't be scaled between 0 and 255. Use "
                         "'tensor.astype(np.uint8)' instead."
                         .format(tensor.min(), tensor.max()))

    # cast tensor
    if parse_version(skimage.__version__) < parse_version("0.16.0"):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            tensor = img_as_ubyte(tensor)
    else:
        tensor = img_as_ubyte(tensor)

    return tensor


def cast_img_uint16(tensor):
    """Cast the data in np.uint16.

    Negative values are not allowed as the skimage method ``img_as_uint`` would
    clip them to 0. Positives values are scaled between 0 and 65535, excepted
    if they fit directly in 16 bit (in this case values are not modified).

    Parameters
    ----------
    tensor : np.ndarray
        Image to cast.

    Returns
    -------
    tensor : np.ndarray, np.uint16
        Image cast.

    """
    # check tensor dtype
    check_array(
        tensor,
        ndim=[2, 3, 4, 5],
        dtype=[np.uint8, np.uint16, np.uint32, np.uint64,
               np.int8, np.int16, np.int32, np.int64,
               np.float32, np.float64])
    if tensor.dtype in [np.float32, np.float64]:
        check_range_value(tensor, min_=0, max_=1)
    elif tensor.dtype in [np.int8, np.int16, np.int32, np.int64]:
        check_range_value(tensor, min_=0)

    if tensor.dtype == np.uint16:
        return tensor

    if (tensor.dtype in [np.uint32, np.uint64, np.int32, np.int64]
            and tensor.max() <= 65535):
        raise ValueError("Tensor values are between {0} and {1}. It fits in "
                         "16 bits and won't be scaled between 0 and 65535. "
                         "Use 'tensor.astype(np.uint16)' instead."
                         .format(tensor.min(), tensor.max()))

    # cast tensor
    if parse_version(skimage.__version__) < parse_version("0.16.0"):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            tensor = img_as_uint(tensor)
    else:
        tensor = img_as_uint(tensor)

    return tensor


def cast_img_float32(tensor):
    """Cast the data in np.float32.

    If the input data is in (unsigned) integer, the values are scaled between
    0 and 1. When converting from a np.float dtype, values are not modified.

    Parameters
    ----------
    tensor : np.ndarray
        Image to cast.

    Returns
    -------
    tensor : np.ndarray, np.float32
        image cast.

    """
    # check tensor dtype
    check_array(
        tensor,
        ndim=[2, 3, 4, 5],
        dtype=[np.uint8, np.uint16, np.uint32, np.uint64,
               np.int8, np.int16, np.int32, np.int64,
               np.float32, np.float64])

    # cast tensor
    if parse_version(skimage.__version__) < parse_version("0.16.0"):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            tensor = img_as_float32(tensor)
    else:
        tensor = img_as_float32(tensor)

    return tensor


def cast_img_float64(tensor):
    """Cast the data in np.float64.

    If the input data is in (unsigned) integer, the values are scaled between
    0 and 1. When converting from a np.float dtype, values are not modified.

    Parameters
    ----------
    tensor : np.ndarray
        Tensor to cast.

    Returns
    -------
    tensor : np.ndarray, np.float64
        Tensor cast.

    """
    # check tensor dtype
    check_array(
        tensor,
        ndim=[2, 3, 4, 5],
        dtype=[np.uint8, np.uint16, np.uint32, np.uint64,
               np.int8, np.int16, np.int32, np.int64,
               np.float32, np.float64])

    # cast tensor
    tensor = img_as_float64(tensor)

    return tensor


# ### Format and crop images ###

def resize_image(image, output_shape, method="bilinear"):
    """Resize an image with bilinear interpolation or nearest neighbor method.

    Parameters
    ----------
    image : np.ndarray
        Image to resize.
    output_shape : Tuple[int]
        Shape of the resized image.
    method : str
        Interpolation method to use.

    Returns
    -------
    image_resized : np.ndarray
        Resized image.

    """
    # check parameters
    check_parameter(output_shape=tuple, method=str)
    check_array(image, ndim=[2, 3], dtype=[np.uint8, np.uint16, np.float32])

    # resize image
    if method == "bilinear":
        image_resized = resize(
            image,
            output_shape,
            mode="reflect",
            preserve_range=True,
            order=1,
            anti_aliasing=True)
    elif method == "nearest":
        image_resized = resize(
            image,
            output_shape,
            mode="reflect",
            preserve_range=True,
            order=0,
            anti_aliasing=False)
    else:
        raise ValueError("Method {0} is not available. Choose between "
                         "'bilinear' or 'nearest' instead.".format(method))

    # cast output dtype
    image_resized = image_resized.astype(image.dtype)

    return image_resized


def get_marge_padding(height, width, x):
    """Pad image to make its shape a multiple of `x`.

    Parameters
    ----------
    height : int
        Original height of the image.
    width : int
        Original width of the image.
    x : int
        Padded image have a `height` and `width` multiple of `x`.

    Returns
    -------
    marge_padding : List[List]
        List of lists with the format
        [[`marge_height_t`, `marge_height_b`], [`marge_width_l`,
        `marge_width_r`]].

    """
    # check parameters
    check_parameter(height=int, width=int, x=int)

    # pad height and width to make it multiple of x
    marge_sup_height = x - (height % x)
    marge_sup_height_l = int(marge_sup_height / 2)
    marge_sup_height_r = marge_sup_height - marge_sup_height_l
    marge_sup_width = x - (width % x)
    marge_sup_width_l = int(marge_sup_width / 2)
    marge_sup_width_r = marge_sup_width - marge_sup_width_l
    marge_padding = [[marge_sup_height_l, marge_sup_height_r],
                     [marge_sup_width_l, marge_sup_width_r]]

    return marge_padding

def unmix_channels(image, number_channels = 2):
    """Unmixes interlaced image stack 

    Parameters
    ----------
    image : np.ndarray
        Image to resize.
    number_channels : int
        number of interlaced channels

    Returns
    -------
    unmixed_images : list of np.ndarray
        Unmixed images.
    """

    
    im_size = image.shape
    number_z_planes = int(im_size[0]/number_channels)

    output_images= [np.zeros((number_z_planes,im_size[1],im_size[2])) for x in range(number_channels)]

    for i in range(number_channels):
        for j in range(number_z_planes):
            output_images[i][j,:,:] = image[2*j+i,:,:]

    return output_images